import contextlib
import json
import re
import threading

from queue import Empty, Queue
from typing import (
    Any,
    Collection,
    Dict,
    List,
    Mapping,
    Optional,
    Tuple,
    Type,
    Union,
)

import paho.mqtt.client as mqtt

from platypush.entities import (
    DimmerEntityManager,
    Entity,
    EnumSwitchEntityManager,
    LightEntityManager,
    SensorEntityManager,
    SwitchEntityManager,
)
from platypush.entities.batteries import Battery
from platypush.entities.devices import Device
from platypush.entities.dimmers import Dimmer
from platypush.entities.electricity import (
    CurrentSensor,
    EnergySensor,
    PowerSensor,
    VoltageSensor,
)
from platypush.entities.humidity import HumiditySensor
from platypush.entities.illuminance import IlluminanceSensor
from platypush.entities.lights import Light
from platypush.entities.linkquality import LinkQuality
from platypush.entities.sensors import (
    BinarySensor,
    EnumSensor,
    NumericSensor,
    Sensor,
)
from platypush.entities.switches import Switch, EnumSwitch
from platypush.entities.temperature import TemperatureSensor
from platypush.message.event.zigbee.mqtt import (
    ZigbeeMqttOnlineEvent,
    ZigbeeMqttOfflineEvent,
    ZigbeeMqttDevicePairingEvent,
    ZigbeeMqttDeviceConnectedEvent,
    ZigbeeMqttDeviceBannedEvent,
    ZigbeeMqttDeviceRemovedEvent,
    ZigbeeMqttDeviceRemovedFailedEvent,
    ZigbeeMqttDeviceWhitelistedEvent,
    ZigbeeMqttDeviceRenamedEvent,
    ZigbeeMqttDeviceBindEvent,
    ZigbeeMqttDevicePropertySetEvent,
    ZigbeeMqttDeviceUnbindEvent,
    ZigbeeMqttGroupAddedEvent,
    ZigbeeMqttGroupAddedFailedEvent,
    ZigbeeMqttGroupRemovedEvent,
    ZigbeeMqttGroupRemovedFailedEvent,
    ZigbeeMqttGroupRemoveAllEvent,
    ZigbeeMqttGroupRemoveAllFailedEvent,
    ZigbeeMqttErrorEvent,
)
from platypush.message.response import Response
from platypush.plugins.mqtt import DEFAULT_TIMEOUT, MqttClient, MqttPlugin, action
from ._state import BridgeState, ZigbeeState


# pylint: disable=too-many-ancestors
class ZigbeeMqttPlugin(
    MqttPlugin,
    DimmerEntityManager,
    EnumSwitchEntityManager,
    LightEntityManager,
    SensorEntityManager,
    SwitchEntityManager,
):
    """
    Support for Zigbee devices using any Zigbee adapter compatible with
    `zigbee2mqtt <https://www.zigbee2mqtt.io/>`_.

    In order to get started you'll need:

        - A Zigbee USB adapter/sniffer (in this example I'll use the
          `CC2531 <https://hackaday.io/project/163487-zigbee-cc2531-smart-home-usb-adapter>`_.
        - A Zigbee debugger/emulator + downloader cable (only to flash the firmware).

    Instructions:

        - Install `cc-tool <https://github.com/dashesy/cc-tool>`_ either from
          sources or from a package manager.
        - Connect the Zigbee to your PC/RaspberryPi in this way: ::

            USB -> CC debugger -> downloader cable -> CC2531 -> USB

        - The debugger and the adapter should be connected *at the same time*.
          If the later ``cc-tool`` command throws up an error, put the device in
          sync while connected by pressing the _Reset_ button on the debugger.
        - Check where the device is mapped. On Linux it will usually be ``/dev/ttyACM0``.
        - Download the latest `Z-Stack firmware
          <https://github.com/Koenkk/Z-Stack-firmware/tree/master/coordinator>`_
          to your device. Instructions for a CC2531 device:

          .. code-block:: shell

             # Check out the latest version of the coordinator firmware at
             # https://github.com/Koenkk/Z-Stack-firmware/tree/master/coordinator

             wget https://github.com/Koenkk/Z-Stack-firmware/raw/master/coordinator/<dir>/bin/default/<version>.zip
             unzip CC2531_DEFAULT_20201127.zip
             [sudo] cc-tool -e -w CC2531ZNP-Prod.hex

        - You can disconnect your debugger and downloader cable once the firmware is flashed.

        - Install ``zigbee2mqtt``. First install a node/npm environment, then
          either install ``zigbee2mqtt`` manually or through your package
          manager. **NOTE**: many API breaking changes have occurred on
          Zigbee2MQTT 1.17.0, therefore this integration will only be compatible
          with the version 1.17.0 of the service or higher versions. Manual
          instructions:

          .. code-block:: shell

             # Clone zigbee2mqtt repository
             export ZIGBEE2MQTT_DIR="$HOME/zigbee2mqtt"
             git clone https://github.com/Koenkk/zigbee2mqtt.git "$ZIGBEE2MQTT_DIR"
             cd "$ZIGBEE2MQTT_DIR"
             # Install dependencies
             npm install

        - You need to have an MQTT broker running somewhere. If not, you can
          install `Mosquitto <https://mosquitto.org/>`_ through your package
          manager on any device in your network.

        - Edit ``$ZIGBEE2MQTT_DIR/data/configuration.yaml`` file to match the configuration of
          your MQTT broker:

          .. code-block:: yaml

             # MQTT settings
             mqtt:
                 # MQTT base topic for zigbee2mqtt MQTT messages
                 topic_prefix: zigbee2mqtt
                 # MQTT server URL
                 server: 'mqtt://localhost'
                 # MQTT server authentication, uncomment if required:
                 # user: my_user
                 # password: my_password

        - Also make sure that ``permit_join`` is set to ``True``, in order to
          allow Zigbee devices to join the network while you're configuring it.
          It's equally important to set ``permit_join`` to ``False`` once you
          have configured your network, to prevent accidental/malignant joins
          from outer Zigbee devices.

        - Start the ``zigbee2mqtt`` daemon on your device (the `official
          documentation
          <https://www.zigbee2mqtt.io/getting_started/running_zigbee2mqtt.html#5-optional-running-as-a-daemon-with-systemctl>`_
          also contains instructions on how to configure it as a ``systemd``
          service:

          .. code-block:: shell

              cd "$ZIGBEE2MQTT_DIR"
              npm start

        - If you have Zigbee devices that are paired to other bridges, unlink
          them or do a factory reset to pair them to your new bridge.

        - If it all goes fine, once the daemon is running and a new device is
          found you should see traces like this in the output of
          ``zigbee2mqtt``::

            zigbee2mqtt:info  2019-11-09T12:19:56: Successfully interviewed '0x00158d0001dc126a',
            device has successfully been paired

        - You are now ready to use this integration.

    """  # noqa: E501

    def __init__(
        self,
        host: str,
        port: int = 1883,
        topic_prefix: str = 'zigbee2mqtt',
        base_topic: Optional[str] = None,
        timeout: int = 10,
        tls_certfile: Optional[str] = None,
        tls_keyfile: Optional[str] = None,
        tls_version: Optional[str] = None,
        tls_ciphers: Optional[str] = None,
        username: Optional[str] = None,
        password: Optional[str] = None,
        **kwargs,
    ):
        """
        :param host: Default MQTT broker where ``zigbee2mqtt`` publishes its messages.
        :param port: Broker listen port (default: 1883).
        :param topic_prefix: Prefix for the published topics, as specified in
            ``ZIGBEE2MQTT_DIR/data/configuration.yaml`` (default: '``zigbee2mqtt``').
        :param base_topic: Legacy alias for ``topic_prefix`` (default:
            '``zigbee2mqtt``').
        :param timeout: If the command expects from a response, then this
            timeout value will be used (default: 60 seconds).
        :param tls_cafile: If the connection requires TLS/SSL, specify the
            certificate authority file (default: None)
        :param tls_certfile: If the connection requires TLS/SSL, specify the
            certificate file (default: None)
        :param tls_keyfile: If the connection requires TLS/SSL, specify the key
            file (default: None)
        :param tls_version: If the connection requires TLS/SSL, specify the
            minimum TLS supported version (default: None)
        :param tls_ciphers: If the connection requires TLS/SSL, specify the
            supported ciphers (default: None)
        :param username: If the connection requires user authentication, specify
            the username (default: None)
        :param password: If the connection requires user authentication, specify
            the password (default: None)
        """
        if base_topic:
            self.logger.warning(
                'base_topic is deprecated, please use topic_prefix instead'
            )
            topic_prefix = base_topic

        super().__init__(
            host=host,
            port=port,
            topics=[
                f'{topic_prefix}/{topic}'
                for topic in [
                    'bridge/state',
                    'bridge/log',
                    'bridge/logging',
                    'bridge/devices',
                    'bridge/groups',
                ]
            ],
            tls_certfile=tls_certfile,
            tls_keyfile=tls_keyfile,
            tls_version=tls_version,
            tls_ciphers=tls_ciphers,
            username=username,
            password=password,
            timeout=timeout,
            **kwargs,
        )

        self.topic_prefix = topic_prefix
        self._info = ZigbeeState()

    @staticmethod
    def _get_properties(device: dict) -> dict:
        """
        Static method that parses the properties of a device from its received
        definition.
        """
        exposes = (device.get('definition') or {}).get('exposes', []).copy()
        properties = {}

        while exposes:
            exposed = exposes.pop(0)
            exposes += exposed.get('features', [])
            if exposed.get('property'):
                properties[exposed['property']] = exposed

        return properties

    @staticmethod
    def _get_options(device: dict) -> dict:
        """
        Static method that parses the options of a device from its received
        definition.
        """
        return {
            option['property']: option
            for option in (device.get('definition') or {}).get('options', [])
            if option.get('property')
        }

    def transform_entities(self, entities: Collection[dict]) -> List[Entity]:
        compatible_entities = []
        for dev in entities:
            if not dev:
                continue

            dev_def = dev.get('definition') or {}
            dev_info = {
                attr: dev.get(attr)
                for attr in (
                    'type',
                    'date_code',
                    'ieee_address',
                    'network_address',
                    'power_source',
                    'software_build_id',
                    'model_id',
                    'model',
                    'vendor',
                    'supported',
                )
            }

            exposed = self._get_properties(dev)
            options = self._get_options(dev)
            reachable = dev.get('supported', False) and not dev.get(
                'interviewing', False
            )

            light_info = self._get_light_meta(dev)
            dev_entities: List[Entity] = [
                *self._get_sensors(dev, exposed, options),
                *self._get_dimmers(dev, exposed, options),
                *self._get_switches(dev, exposed, options),
                *self._get_enum_switches(dev, exposed, options),
            ]

            if light_info:
                dev_entities.append(
                    Light(
                        id=f'{dev["ieee_address"]}:light',
                        name='Light',
                        on=dev.get('state', {}).get('state')
                        == light_info.get('value_on'),
                        brightness_min=light_info.get('brightness_min'),
                        brightness_max=light_info.get('brightness_max'),
                        temperature_min=light_info.get('temperature_min'),
                        temperature_max=light_info.get('temperature_max'),
                        hue_min=light_info.get('hue_min'),
                        hue_max=light_info.get('hue_max'),
                        saturation_min=light_info.get('saturation_min'),
                        saturation_max=light_info.get('saturation_max'),
                        brightness=(
                            dev.get('state', {}).get(
                                light_info.get('brightness_name', 'brightness')
                            )
                        ),
                        temperature=(
                            dev.get('state', {}).get(
                                light_info.get('temperature_name', 'temperature')
                            )
                        ),
                        hue=(
                            dev.get('state', {})
                            .get('color', {})
                            .get(light_info.get('hue_name', 'hue'))
                        ),
                        saturation=(
                            dev.get('state', {})
                            .get('color', {})
                            .get(light_info.get('saturation_name', 'saturation'))
                        ),
                        x=(
                            dev.get('state', {})
                            .get('color', {})
                            .get(light_info.get('x_name', 'x'))
                        ),
                        y=(
                            dev.get('state', {})
                            .get('color', {})
                            .get(light_info.get('y_name', 'y'))
                        ),
                        description=dev_def.get('description'),
                        data=dev_info,
                    )
                )

            if dev_entities:
                parent = Device(
                    id=dev['ieee_address'],
                    name=dev.get('friendly_name'),
                    description=dev_def.get('description'),
                    external_url=self._get_device_url(dev),
                    image_url=self._get_image_url(dev),
                    reachable=reachable,
                )

                for entity in dev_entities:
                    entity.parent = parent
                dev_entities.append(parent)

            compatible_entities += dev_entities

        return compatible_entities

    @staticmethod
    def _get_device_url(device_info: dict) -> Optional[str]:
        """
        Static method that returns the zigbee2mqtt URL with the information
        about a certain device, if the model is available in its definition.
        """
        model = device_info.get('definition', {}).get('model')
        if not model:
            return None

        return f'https://www.zigbee2mqtt.io/devices/{model}.html'

    @staticmethod
    def _get_image_url(device_info: dict) -> Optional[str]:
        """
        Static method that returns the zigbee2mqtt URL of the image of a
        certain device, if the model is available in its definition.
        """
        model = device_info.get('definition', {}).get('model')
        if not model:
            return None

        return f'https://www.zigbee2mqtt.io/images/devices/{model}.jpg'

    def _get_network_info(self, **kwargs) -> dict:
        """
        Refreshes the network information.
        """
        self.logger.info('Fetching Zigbee network information')
        client = None
        mqtt_args = self._mqtt_args(**kwargs)
        timeout = mqtt_args.pop('timeout', DEFAULT_TIMEOUT)
        info: Dict[str, Any] = {
            'state': None,
            'info': {},
            'config': {},
            'devices': [],
            'groups': [],
        }

        info_ready_events = {topic: threading.Event() for topic in info}

        def msg_callback(_, __, msg):
            topic = msg.topic.split('/')[-1]
            if topic in info:
                info[topic] = (
                    msg.payload.decode()
                    if topic == 'state'
                    else json.loads(msg.payload.decode())
                )
                info_ready_events[topic].set()

        try:
            client = self._get_client(  # pylint: disable=unexpected-keyword-arg
                **mqtt_args
            )
            client.on_message = msg_callback
            client.connect()
            client.subscribe(self.topic_prefix + '/bridge/#')
            client.loop_start()

            for event in info_ready_events.values():
                info_ready = event.wait(timeout=timeout)
                if not info_ready:
                    raise TimeoutError(
                        'A timeout occurred while fetching the Zigbee network information'
                    )

            # Cache the new results
            self._info.devices.by_name = {
                self._preferred_name(device): device
                for device in info.get('devices', [])
            }

            self._info.devices.by_address = {
                device['ieee_address']: device for device in info.get('devices', [])
            }

            self._info.groups = {
                group.get('name'): group for group in info.get('groups', [])
            }

            self.logger.info('Zigbee network configuration updated')
        finally:
            try:
                if client:
                    client.loop_stop()
                    client.disconnect()
            except Exception as e:
                self.logger.warning('Error on MQTT client disconnection: %s', e)

        return info

    def _topic(self, topic: str) -> str:
        """
        Utility method that construct a topic prefixed by the configured base
        topic.
        """
        return f'{self.topic_prefix}/{topic}'

    @staticmethod
    def _parse_response(response: Union[dict, Response]) -> dict:
        """
        Utility method that flattens a response received on a zigbee2mqtt topic
        into a dictionary.
        """
        if isinstance(response, Response):
            rs: dict = response.output  # type: ignore
            response = rs

        if isinstance(response, dict):
            assert response.get('status') != 'error', response.get(
                'error', 'zigbee2mqtt error'
            )

        return response or {}

    def _run_request(
        self,
        topic: str,
        msg: Union[dict, str],
        reply_topic: Optional[str] = None,
        **kwargs,
    ) -> dict:
        """
        Sends a request/message to the Zigbee2MQTT bridge and waits for a
        response.
        """
        return self._parse_response(
            self.publish(  # type: ignore
                topic=topic,
                msg=msg,
                reply_topic=reply_topic,
                **self._mqtt_args(**kwargs),
            )
            or {}
        )

    @action
    def devices(self, **kwargs) -> List[Dict[str, Any]]:
        """
        Get the list of devices registered to the service.

        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).

        :return: List of paired devices. Example output:

        .. code-block:: json

            [
                {
                    "date_code": "20190608",
                    "friendly_name": "Coordinator",
                    "ieee_address": "0x00123456789abcde",
                    "network_address": 0,
                    "supported": false,
                    "type": "Coordinator",
                    "interviewing": false,
                    "interviewing_completed": true,
                    "definition": null,
                    "endpoints": {
                        "13": {
                            "bindings": [],
                            "clusters": {
                                "input": ["genOta"],
                                "output": []
                            },
                            "output": []
                        }
                    }
                },

                {
                    "date_code": "20180906",
                    "friendly_name": "My Light Bulb",
                    "ieee_address": "0x00123456789abcdf",
                    "network_address": 52715,
                    "power_source": "Mains (single phase)",
                    "software_build_id": "5.127.1.26581",
                    "model_id": "LCT001",
                    "supported": true,
                    "interviewing": false,
                    "interviewing_completed": true,
                    "type": "Router",
                    "definition": {
                        "description": "Hue white and color ambiance E26/E27/E14",
                        "model": "9290012573A",
                        "vendor": "Philips",
                        "exposes": [
                            {
                                "features": [
                                    {
                                        "access": 7,
                                        "description": "On/off state of this light",
                                        "name": "state",
                                        "property": "state",
                                        "type": "binary",
                                        "value_off": "OFF",
                                        "value_on": "ON",
                                        "value_toggle": "TOGGLE"
                                    },
                                    {
                                        "access": 7,
                                        "description": "Brightness of this light",
                                        "name": "brightness",
                                        "property": "brightness",
                                        "type": "numeric",
                                        "value_max": 254,
                                        "value_min": 0
                                    },
                                    {
                                        "access": 7,
                                        "description": "Color temperature of this light",
                                        "name": "color_temp",
                                        "property": "color_temp",
                                        "type": "numeric",
                                        "unit": "mired",
                                        "value_max": 500,
                                        "value_min": 150
                                    },
                                    {
                                        "description": "Color of this light in the XY space",
                                        "features": [
                                            {
                                                "access": 7,
                                                "name": "x",
                                                "property": "x",
                                                "type": "numeric"
                                            },
                                            {
                                                "access": 7,
                                                "name": "y",
                                                "property": "y",
                                                "type": "numeric"
                                            }
                                        ],
                                        "name": "color_xy",
                                        "property": "color",
                                        "type": "composite"
                                    }
                                ],
                                "type": "light"
                            },
                            {
                                "access": 2,
                                "description": "Triggers an effect on the light",
                                "name": "effect",
                                "property": "effect",
                                "type": "enum",
                                "values": [
                                    "blink",
                                    "breathe",
                                    "okay",
                                    "channel_change",
                                    "finish_effect",
                                    "stop_effect"
                                ]
                            },
                            {
                                "access": 1,
                                "description": "Link quality (signal strength)",
                                "name": "linkquality",
                                "property": "linkquality",
                                "type": "numeric",
                                "unit": "lqi",
                                "value_max": 255,
                                "value_min": 0
                            }
                        ]
                    },

                    "endpoints": {
                        "11": {
                            "bindings": [],
                            "clusters": {
                                "input": [
                                    "genBasic",
                                    "genIdentify",
                                    "genGroups",
                                    "genScenes",
                                    "genOnOff",
                                    "genLevelCtrl",
                                    "touchlink",
                                    "lightingColorCtrl",
                                ],
                                "output": [
                                    "genOta"
                                ]
                            },
                            "configured_reportings": []
                        },
                        "242": {
                            "bindings": [],
                            "clusters": {
                                "input": [
                                    "greenPower"
                                ],
                                "output": [
                                    "greenPower"
                                ]
                            },
                            "configured_reportings": []
                        }
                    }
                }
            ]

        """
        return self._get_network_info(**kwargs).get('devices', {})

    @action
    def permit_join(
        self, permit: bool = True, timeout: Optional[float] = None, **kwargs
    ):
        """
        Enable/disable devices from joining the network.

        This is not persistent (it will not be saved to ``configuration.yaml``).

        :param permit: Set to True to allow joins, False otherwise.
        :param timeout: Allow/disallow joins only for this amount of time.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        if timeout:
            return self._run_request(
                topic=self._topic('bridge/request/permit_join'),
                msg={'value': permit, 'time': timeout},
                reply_topic=self._topic('bridge/response/permit_join'),
                **self._mqtt_args(**kwargs),
            )

        return self.publish(
            topic=self._topic('bridge/request/permit_join'),
            msg={'value': permit},
            **self._mqtt_args(**kwargs),
        )

    @action
    def factory_reset(self, **kwargs):
        """
        Perform a factory reset of a device connected to the network, following
        the procedure required by the particular device (for instance, Hue bulbs
        require the Zigbee adapter to be close to the device while a button on
        the back of the bulb is pressed).

        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        self.publish(
            topic=self._topic('bridge/request/touchlink/factory_reset'),
            msg='',
            **self._mqtt_args(**kwargs),
        )

    @action
    def log_level(self, level: str, **kwargs):
        """
        Change the log level at runtime.

        This change will not be persistent.

        :param level: Possible values: 'debug', 'info', 'warn', 'error'.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/config/log_level'),
            msg={'value': level},
            reply_topic=self._topic('bridge/response/config/log_level'),
            **self._mqtt_args(**kwargs),
        )

    @action
    def device_set_option(self, device: str, option: str, value: Any, **kwargs):
        """
        Change the options of a device.

        Options can only be changed, not added or deleted.

        :param device: Display name or IEEE address of the device.
        :param option: Option name.
        :param value: New value.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/options'),
            reply_topic=self._topic('bridge/response/device/options'),
            msg={
                'id': device,
                'options': {
                    option: value,
                },
            },
            **self._mqtt_args(**kwargs),
        )

    @action
    def device_remove(self, device: str, force: bool = False, **kwargs):
        """
        Remove a device from the network.

        :param device: Display name of the device.
        :param force: Force the remove also if the removal wasn't acknowledged
            by the device. Note: a forced remove only removes the entry from the
            internal database, but the device is likely to connect again when
            restarted unless it's factory reset (default: False).
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/remove'),
            msg={'id': device, 'force': force},
            reply_topic=self._topic('bridge/response/device/remove'),
            **self._mqtt_args(**kwargs),
        )

    @action
    def device_ban(self, device: str, **kwargs):
        """
        Ban a device from the network.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/ban'),
            reply_topic=self._topic('bridge/response/device/ban'),
            msg={'id': device},
            **self._mqtt_args(**kwargs),
        )

    @action
    def device_whitelist(self, device: str, **kwargs):
        """
        Whitelist a device on the network.

        Note: once at least a device is whitelisted, all the other
        non-whitelisted devices will be removed from the network.

        :param device: Display name of the device.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/whitelist'),
            reply_topic=self._topic('bridge/response/device/whitelist'),
            msg={'id': device},
            **self._mqtt_args(**kwargs),
        )

    @action
    def device_rename(self, name: str, device: Optional[str] = None, **kwargs):
        """
        Rename a device on the network.

        :param name: New name.
        :param device: Current name of the device to rename. If no name is
            specified then the rename will affect the last device that joined
            the network.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        if name == device:
            self.logger.info('Old and new name are the same: nothing to do')
            return None

        devices: dict = self.devices().output  # type: ignore
        assert not [
            dev for dev in devices if dev.get('friendly_name') == name
        ], f'A device named {name} already exists on the network'

        if device:
            req: Dict[str, Any] = {
                'from': device,
                'to': name,
            }
        else:
            req = {
                'last': True,
                'to': name,
            }

        return self._run_request(
            topic=self._topic('bridge/request/device/rename'),
            msg=req,
            reply_topic=self._topic('bridge/response/device/rename'),
            **self._mqtt_args(**kwargs),
        )

    @staticmethod
    def _build_device_get_request(values: List[Dict[str, Any]]) -> Dict[str, Any]:
        """
        Prepares a ``device_get`` request, as a dictionary to be sent down to
        the bridge.

        This makes sure that the properties and options are properly mapped and
        converted.
        """

        def extract_value(val: dict, root: dict, depth: int = 0):
            for feature in val.get('features', []):
                new_root = root
                if depth > 0:
                    new_root = root[val['property']] = root.get(val['property'], {})

                extract_value(feature, new_root, depth=depth + 1)

            if not val.get('access', 1) & 0x4:
                # Property not readable/query-able
                return

            if 'features' not in val:
                if 'property' in val:
                    root[val['property']] = 0 if val['type'] == 'numeric' else ''
                return

            if 'property' in val:
                root[val['property']] = root.get(val['property'], {})

        ret: Dict[str, Any] = {}
        for value in values:
            extract_value(value, root=ret)

        return ret

    def _get_device_info(self, device: str, **kwargs) -> dict:
        """
        Get or retrieve the information about a device.
        """
        device_info = self._info.devices.by_name.get(
            # First: check by friendly name
            device,
            # Second: check by address
            self._info.devices.by_address.get(device, {}),
        )

        if not device_info:
            # Third: try and get the device from upstream
            network_info = self._get_network_info(**kwargs)
            next(
                iter(
                    d
                    for d in network_info.get('devices', [])
                    if self._device_name_matches(device, d)
                ),
                {},
            )

        return device_info

    @staticmethod
    def _preferred_name(device: dict) -> str:
        """
        Utility method that returns the preferred name of a device, on the basis
        of which attributes are exposed (friendly name or IEEE address).
        """
        return device.get('friendly_name') or device.get('ieee_address') or ''

    @classmethod
    def _device_name_matches(cls, name: str, device: dict) -> bool:
        """
        Utility method that checks if either the friendly name or IEEE address
        of a device match a certain string.
        """
        name = str(cls._ieee_address(name))
        return name == device.get('friendly_name') or name == device.get('ieee_address')

    @action
    def device_get(  # pylint: disable=redefined-builtin
        self, device: str, property: Optional[str] = None, **kwargs
    ) -> Dict[str, Any]:
        """
        Get the properties of a device.

        The returned keys vary depending on the device. For example, a light
        bulb may have the "``state``" and "``brightness``" properties, while an
        environment sensor may have the "``temperature``" and "``humidity``"
        properties, and so on.

        :param device: Display name of the device.
        :param property: Name of the property that should be retrieved (default:
            all).
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        :return: Key->value map of the device properties.
        """
        kwargs = self._mqtt_args(**kwargs)
        device_info = self._get_device_info(device, **kwargs)
        assert device_info, f'No such device: {device}'
        device = self._preferred_name(device_info)

        if property:
            properties = self._run_request(
                topic=self._topic(device) + f'/get/{property}',
                reply_topic=self._topic(device),
                msg={property: ''},
                **kwargs,
            )

            assert property in properties, f'No such property: {property}'
            return {property: properties[property]}

        exposes = (device_info.get('definition', {}) or {}).get('exposes', [])
        if not exposes:
            return {}

        # If the device has no queryable properties, don't specify a reply
        # topic to listen on
        req = self._build_device_get_request(exposes)
        reply_topic = self._topic(device) if req else None
        return self._run_request(
            topic=self._topic(device) + '/get',
            reply_topic=reply_topic,
            msg=req,
            **kwargs,
        )

    @action
    def devices_get(
        self, devices: Optional[List[str]] = None, **kwargs
    ) -> Dict[str, dict]:
        """
        Get the properties of the devices connected to the network.

        :param devices: If set, then only the status of these devices (by
            friendly name) will be retrieved (default: retrieve all).
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        :return: Key->value map of the device properties:

            .. code-block:: json

                {
                    "Bulb": {
                        "state": "ON",
                        "brightness": 254
                    },
                    "Sensor": {
                        "temperature": 22.5
                    }
                }
        """
        kwargs = self._mqtt_args(**kwargs)

        if not devices:
            devices = list(
                {
                    self._preferred_name(device)
                    for device in list(self.devices(**kwargs).output or [])
                    if self._preferred_name(device)
                }
            )

        def worker(device: str, q: Queue):
            q.put_nowait(self.device_get(device, **kwargs).output)  # type: ignore

        queues: Dict[str, Queue] = {}
        workers = {}
        response = {}

        for device in devices:
            queues[device] = Queue()
            workers[device] = threading.Thread(
                target=worker, args=(device, queues[device])
            )
            workers[device].start()

        for device in devices:
            timeout = kwargs.get('timeout')
            try:
                response[device] = queues[device].get(timeout=timeout)
                workers[device].join(timeout=timeout)
            except Empty:
                self.logger.warning(
                    'Could not get the status of the device %s: timeout after %f seconds',
                    device,
                    timeout,
                )
            except Exception as e:
                self.logger.warning(
                    'An error occurred while getting the status of the device %s: %s',
                    device,
                    e,
                )

                self.logger.exception(e)

        return response

    @action
    def status(self, *args, device: Optional[str] = None, **kwargs):
        """
        Get the status of a device (by friendly name) or of all the connected
        devices (it wraps :meth:`.devices_get`).

        :param device: Device friendly name (default: get all devices).
        """
        return self.devices_get([device] if device else None, *args, **kwargs)

    @action
    def device_set(  # pylint: disable=redefined-builtin
        self,
        device: str,
        property: Optional[str] = None,
        value: Optional[Any] = None,
        values: Optional[Dict[str, Any]] = None,
        **kwargs,
    ):
        """
        Set a properties on a device.

        The compatible properties vary depending on the device. For example, a
        light bulb may have the "``state``" and "``brightness``" properties,
        while an environment sensor may have the "``temperature``" and
        "``humidity``" properties, and so on.

        :param device: Display name of the device.
        :param property: Name of the property that should be set.
        :param value: New value of the property.
        :param values: If you want to set multiple values, then pass this
            mapping instead of ``property``+``value``.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        msg = (values or {}).copy()
        reply_topic = None
        device_info = self._get_device_info(device, **kwargs)
        assert device_info, f'No such device: {device}'
        device = self._preferred_name(device_info)

        if property:
            # Check if we're trying to set an option
            stored_option = self._get_options(device_info).get(property)
            if stored_option:
                return self.device_set_option(device, property, value, **kwargs)

            # Check if it's a property
            reply_topic = self._topic(device)
            stored_property = self._get_properties(device_info).get(property)
            assert stored_property, f'No such property: {property}'

            # Set the new value on the message
            msg[property] = value

            # Don't wait for an update from a value that is not readable
            if self._is_write_only(stored_property):
                reply_topic = None

        properties = self._run_request(
            topic=self._topic(device + '/set'),
            reply_topic=reply_topic,
            msg=msg,
            **self._mqtt_args(**kwargs),
        )

        if property and reply_topic:
            assert (
                property in properties
            ), f'Could not retrieve the new state for {property}'
            return {property: properties[property]}

        return properties

    @action
    # pylint: disable=redefined-builtin
    def set_value(
        self, device: str, property: Optional[str] = None, data=None, **kwargs
    ):
        """
        Entity-compatible way of setting a value on a node.

        :param device: Device friendly name, IEEE address or internal entity ID
            in ``<address>:<property>`` format.
        :param property: Name of the property to set. If not specified here, it
            should be specified on ``device`` in ``<address>:<property>``
            format.
        :param data: Value to set for the property.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        dev, prop = self._ieee_address_and_property(device)
        if not property:
            property = prop

        self.device_set(dev, property, data, **kwargs)

    @action
    def set(self, entity: str, value: Any, attribute: Optional[str] = None, **kwargs):
        return self.set_value(entity, data=value, property=attribute, **kwargs)

    @action
    def device_check_ota_updates(self, device: str, **kwargs) -> dict:
        """
        Check if the specified device has any OTA updates available to install.

        :param device: Address or friendly name of the device.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).

        :return:

            .. code-block:: json

                {
                    "id": "<device ID>",
                    "update_available": true,
                    "status": "ok"
                }

        """
        ret = self._run_request(
            topic=self._topic('bridge/request/device/ota_update/check'),
            reply_topic=self._topic('bridge/response/device/ota_update/check'),
            msg={'id': device},
            **self._mqtt_args(**kwargs),
        )

        return {
            'status': ret['status'],
            'id': ret.get('data', {}).get('id'),
            'update_available': ret.get('data', {}).get('update_available', False),
        }

    @action
    def device_install_ota_updates(self, device: str, **kwargs):
        """
        Install OTA updates for a device if available.

        :param device: Address or friendly name of the device.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/ota_update/update'),
            reply_topic=self._topic('bridge/response/device/ota_update/update'),
            msg={'id': device},
            **self._mqtt_args(**kwargs),
        )

    @action
    def groups(self, **kwargs) -> List[dict]:
        """
        Get the groups registered on the device.

        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._get_network_info(**kwargs).get('groups', [])

    @action
    def info(self, **kwargs) -> dict:
        """
        Get the information, configuration and state of the network.

        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).

        :return: Example:

            .. code-block:: json

                {
                    "state": "online",
                    "commit": "07cdc9d",
                    "config": {
                        "advanced": {
                            "adapter_concurrent": null,
                            "adapter_delay": null,
                            "availability_blacklist": [],
                            "availability_blocklist": [],
                            "availability_passlist": [],
                            "availability_timeout": 0,
                            "availability_whitelist": [],
                            "cache_state": true,
                            "cache_state_persistent": true,
                            "cache_state_send_on_startup": true,
                            "channel": 11,
                            "elapsed": false,
                            "ext_pan_id": [
                                221,
                                221,
                                221,
                                221,
                                221,
                                221,
                                221,
                                221
                            ],
                            "homeassistant_discovery_topic": "homeassistant",
                            "homeassistant_legacy_triggers": true,
                            "homeassistant_status_topic": "hass/status",
                            "last_seen": "disable",
                            "legacy_api": true,
                            "log_directory": "/opt/zigbee2mqtt/data/log/%TIMESTAMP%",
                            "log_file": "log.txt",
                            "log_level": "debug",
                            "log_output": [
                                "console",
                                "file"
                            ],
                            "log_rotation": true,
                            "log_syslog": {},
                            "pan_id": 6754,
                            "report": false,
                            "soft_reset_timeout": 0,
                            "timestamp_format": "YYYY-MM-DD HH:mm:ss"
                        },
                        "ban": [],
                        "blocklist": [],
                        "device_options": {},
                        "devices": {
                            "0x00123456789abcdf": {
                                "friendly_name": "My Light Bulb"
                            }
                        },
                        "experimental": {
                            "output": "json"
                        },
                        "external_converters": [],
                        "groups": {},
                        "homeassistant": false,
                        "map_options": {
                            "graphviz": {
                                "colors": {
                                    "fill": {
                                        "coordinator": "#e04e5d",
                                        "enddevice": "#fff8ce",
                                        "router": "#4ea3e0"
                                    },
                                    "font": {
                                        "coordinator": "#ffffff",
                                        "enddevice": "#000000",
                                        "router": "#ffffff"
                                    },
                                    "line": {
                                        "active": "#009900",
                                        "inactive": "#994444"
                                    }
                                }
                            }
                        },
                        "mqtt": {
                            "base_topic": "zigbee2mqtt",
                            "force_disable_retain": false,
                            "include_device_information": false,
                            "server": "mqtt://localhost"
                        },
                        "passlist": [],
                        "permit_join": true,
                        "serial": {
                            "disable_led": false,
                            "port": "/dev/ttyUSB0"
                        },
                        "whitelist": []
                    },
                    "coordinator": {
                        "meta": {
                            "maintrel": 3,
                            "majorrel": 2,
                            "minorrel": 6,
                            "product": 0,
                            "revision": 20190608,
                            "transportrev": 2
                        },
                        "type": "zStack12"
                    },
                    "log_level": "debug",
                    "network": {
                        "channel": 11,
                        "extended_pan_id": "0xdddddddddddddddd",
                        "pan_id": 6754
                    },
                    "permit_join": true,
                    "version": "1.17.0"
                }

        """
        info = self._get_network_info(**kwargs)
        return {
            'state': info.get('state'),
            'info': info.get('info'),
        }

    @action
    def group_add(  # pylint: disable=redefined-builtin
        self, name: str, id: Optional[int] = None, **kwargs
    ):
        """
        Add a new group.

        :param name: Display name of the group.
        :param id: Optional numeric ID (default: auto-generated).
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        payload = (
            name
            if id is None
            else {
                'id': id,
                'friendly_name': name,
            }
        )

        return self._run_request(
            topic=self._topic('bridge/request/group/add'),
            reply_topic=self._topic('bridge/response/group/add'),
            msg=payload,
            **self._mqtt_args(**kwargs),
        )

    @action
    def group_get(  # pylint: disable=redefined-builtin
        self, group: str, property: Optional[str] = None, **kwargs
    ) -> dict:
        """
        Get one or more properties of a group.

        The compatible properties vary depending on the devices on the group.
        For example, a light bulb may have the "``state``" (with values ``"ON"``
        and ``"OFF"``) and "``brightness``" properties, while an environment
        sensor may have the "``temperature``" and "``humidity``" properties, and
        so on.

        :param group: Display name of the group.
        :param property: Name of the property to retrieve (default: all
            available properties)
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        msg = {}
        if property:
            msg = {property: ''}

        properties = self._run_request(
            topic=self._topic(group + '/get'),
            reply_topic=self._topic(group),
            msg=msg,
            **self._mqtt_args(**kwargs),
        )

        if property:
            assert property in properties, f'No such property: {property}'
            return {property: properties[property]}

        return properties

    @action
    def group_set(  # pylint: disable=redefined-builtin
        self, group: str, property: str, value: Any, **kwargs
    ):
        """
        Set a properties on a group.

        The compatible properties vary depending on the devices on the group.

        For example, a light bulb may have the "``state``" (with values ``"ON"``
        and ``"OFF"``) and "``brightness``" properties, while an environment
        sensor may have the "``temperature``" and "``humidity``" properties, and
        so on.

        :param group: Display name of the group.
        :param property: Name of the property that should be set.
        :param value: New value of the property.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        properties = self._run_request(
            topic=self._topic(group + '/set'),
            reply_topic=self._topic(group),
            msg={property: value},
            **self._mqtt_args(**kwargs),
        )

        if property:
            assert property in properties, f'No such property: {property}'
            return {property: properties[property]}

        return properties

    @action
    def group_rename(self, name: str, group: str, **kwargs):
        """
        Rename a group.

        :param name: New name.
        :param group: Current name of the group to rename.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        if name == group:
            self.logger.info('Old and new name are the same: nothing to do')
            return None

        groups = {
            g.get('friendly_name'): g
            for g in dict(self.groups().output)  # type: ignore
        }

        assert name not in groups, f'A group named {name} already exists on the network'

        return self._run_request(
            topic=self._topic('bridge/request/group/rename'),
            reply_topic=self._topic('bridge/response/group/rename'),
            msg={'from': group, 'to': name} if group else name,
            **self._mqtt_args(**kwargs),
        )

    @action
    def group_remove(self, name: str, **kwargs):
        """
        Remove a group.

        :param name: Display name of the group.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/group/remove'),
            reply_topic=self._topic('bridge/response/group/remove'),
            msg=name,
            **self._mqtt_args(**kwargs),
        )

    @action
    def group_add_device(self, group: str, device: str, **kwargs):
        """
        Add a device to a group.

        :param group: Display name of the group.
        :param device: Display name of the device to be added.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/group/members/add'),
            reply_topic=self._topic('bridge/response/group/members/add'),
            msg={
                'group': group,
                'device': device,
            },
            **self._mqtt_args(**kwargs),
        )

    @action
    def group_remove_device(self, group: str, device: Optional[str] = None, **kwargs):
        """
        Remove a device from a group.

        :param group: Display name of the group.
        :param device: Display name of the device to be removed. If none is
            specified then all the devices registered to the specified group
            will be removed.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        remove_suffix = '_all' if device is None else ''
        return self._run_request(
            topic=self._topic(f'bridge/request/group/members/remove{remove_suffix}'),
            reply_topic=self._topic(
                f'bridge/response/group/members/remove{remove_suffix}'
            ),
            msg={
                'group': group,
                'device': device,
            },
            **self._mqtt_args(**kwargs),
        )

    @action
    def bind_devices(self, source: str, target: str, **kwargs):
        """
        Bind two devices.

        Binding makes it possible that devices can directly control each other
        without the intervention of zigbee2mqtt or any home automation software.
        You may want to use this feature to bind for example an IKEA/Philips Hue
        dimmer switch to a light bulb, or a Zigbee remote to a thermostat. Read
        more on the `zigbee2mqtt binding page
        <https://www.zigbee2mqtt.io/information/binding.html>`_.

        :param source: Name of the source device. It can also be a group name,
            although the support is `still experimental
            <https://www.zigbee2mqtt.io/information/binding.html#binding-a-group>`_.
            You can also bind a specific device endpoint - for example
            ``MySensor/temperature``.
        :param target: Name of the target device. You can also bind a specific
            device endpoint - for example ``MyLight/state``.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/bind'),
            reply_topic=self._topic('bridge/response/device/bind'),
            msg={'from': source, 'to': target},
            **self._mqtt_args(**kwargs),
        )

    @action
    def unbind_devices(self, source: str, target: str, **kwargs):
        """
        Remove a binding between two devices.

        :param source: Name of the source device. You can also bind a specific
            device endpoint - for example ``MySensor/temperature``.
        :param target: Name of the target device. You can also bind a specific
            device endpoint - for example ``MyLight/state``.
        :param kwargs: Extra arguments to be passed to
            :meth:`platypush.plugins.mqtt.MqttPlugin.publish` (default: query
            the default configured device).
        """
        return self._run_request(
            topic=self._topic('bridge/request/device/unbind'),
            reply_topic=self._topic('bridge/response/device/unbind'),
            msg={'from': source, 'to': target},
            **self._mqtt_args(**kwargs),
        )

    @action
    def on(self, device, *_, **__):  # pylint: disable=arguments-differ
        """
        Turn on/set to true a switch, a binary property or an option.
        """
        device, prop_info = self._get_switch_info(device)
        return self.device_set(
            device, prop_info['property'], prop_info.get('value_on', 'ON')
        )

    @action
    def off(self, device, *_, **__):  # pylint: disable=arguments-differ
        """
        Turn off/set to false a switch, a binary property or an option.
        """
        device, prop_info = self._get_switch_info(device)
        return self.device_set(
            device, prop_info['property'], prop_info.get('value_off', 'OFF')
        )

    @action
    def toggle(self, device, *_, **__):  # pylint: disable=arguments-differ
        """
        Toggles the state of a switch, a binary property or an option.
        """
        device, prop_info = self._get_switch_info(device)
        prop = prop_info['property']
        device_state: dict = self.device_get(device).output  # type: ignore
        return self.device_set(
            device,
            prop,
            prop_info.get(
                'value_toggle',
                'OFF'
                if device_state.get(prop) == prop_info.get('value_on', 'ON')
                else 'ON',
            ),
        )

    def _get_switch_info(self, name: str) -> Tuple[str, dict]:
        """
        Get the information about a switch or switch-like device by name or
        address.
        """
        name, prop = self._ieee_address_and_property(name)
        if not prop or prop == 'light':
            prop = 'state'

        device_info = self._get_device_info(name)
        assert device_info, f'No such device: {name}'
        name = self._preferred_name(device_info)

        prop_info = self._get_properties(device_info).get(prop)
        option = self._get_options(device_info).get(prop)
        if option:
            return name, option

        assert prop_info, f'No such property on device {name}: {prop}'
        return name, prop_info

    @staticmethod
    def _is_read_only(feature: dict) -> bool:
        """
        Utility method that checks if a feature is read-only on the basis of its
        access flags.
        """
        return bool(feature.get('access', 0) & 2) == 0 and (
            bool(feature.get('access', 0) & 1) == 1
            or bool(feature.get('access', 0) & 4) == 1
        )

    @staticmethod
    def _is_write_only(feature: dict) -> bool:
        """
        Utility method that checks if a feature is write-only on the basis of
        its access flags.
        """
        return bool(feature.get('access', 0) & 2) == 1 and (
            bool(feature.get('access', 0) & 1) == 0
            or bool(feature.get('access', 0) & 4) == 0
        )

    @staticmethod
    def _is_query_disabled(feature: dict) -> bool:
        """
        Utility method that checks if a feature doesn't support programmatic
        querying (i.e. it will only broadcast its state when available) on the
        basis of its access flags.
        """
        return bool(feature.get('access', 0) & 4) == 0

    @staticmethod
    def _ieee_address_and_property(
        device: Union[dict, str]
    ) -> Tuple[str, Optional[str]]:
        """
        Given a device property, as a dictionary containing the full device
        definition or a string containing the device address and property,
        return a tuple in the format ``(device_address, property_name)``.
        """
        # Entity value IDs are stored in the `<address>:<property>`
        # format. Therefore, we need to split by `:` if we want to
        # retrieve the original address.
        if isinstance(device, dict):
            dev = device['ieee_address']
        else:
            dev = device

        # IEEE address + property format
        if re.search(r'^0x[0-9a-fA-F]{16}:', dev):
            parts = dev.split(':')
            return parts[0], parts[1] if len(parts) > 1 else None

        return dev, None

    @classmethod
    def _ieee_address(cls, device: Union[dict, str]) -> str:
        """
        :return: The IEEE address of a device, given its full definition or
        common name.
        """
        return cls._ieee_address_and_property(device)[0]

    @classmethod
    def _get_switches(
        cls, device_info: dict, props: dict, options: dict
    ) -> List[Switch]:
        """
        A utility method that parses the properties of a device that can be
        mapped to switches (or switch-like entities).
        """
        return [
            cls._to_entity(
                Switch,
                device_info,
                prop,
                options=options,
                state=device_info.get('state', {}).get(prop['property'])
                == prop['value_on'],
                data={
                    'value_on': prop['value_on'],
                    'value_off': prop['value_off'],
                    'value_toggle': prop.get('value_toggle'),
                },
            )
            for prop in [*props.values(), *options.values()]
            if (
                prop.get('type') == 'binary'
                and 'value_on' in prop
                and 'value_off' in prop
                and not cls._is_read_only(prop)
            )
        ]

    @classmethod
    def _get_sensors(  # pylint: disable=too-many-branches
        cls, device_info: dict, props: dict, options: dict
    ) -> List[Sensor]:
        """
        A utility method that parses the properties of a device that can be
        mapped to sensors (or sensor-like entities).
        """
        sensors = []
        properties = [
            prop
            for prop in [*props.values(), *options.values()]
            if cls._is_read_only(prop)
        ]

        for prop in properties:
            entity_type = None
            sensor_args = {
                'value': device_info.get('state', {}).get(prop['property']),
            }

            if prop.get('type') == 'numeric':
                sensor_args.update(
                    {
                        'min': prop.get('value_min'),
                        'max': prop.get('value_max'),
                        'unit': prop.get('unit'),
                    }
                )

            if prop.get('property') == 'battery':
                entity_type = Battery
            elif prop.get('property') == 'linkquality':
                entity_type = LinkQuality
            elif prop.get('property') == 'current':
                entity_type = CurrentSensor
            elif prop.get('property') == 'energy':
                entity_type = EnergySensor
            elif prop.get('property') == 'power':
                entity_type = PowerSensor
            elif prop.get('property') == 'voltage':
                entity_type = VoltageSensor
            elif prop.get('property', '').endswith('temperature'):
                entity_type = TemperatureSensor
            elif re.search(r'(humidity|moisture)$', prop.get('property' '')):
                entity_type = HumiditySensor
            elif re.search(r'(illuminance|luminosity)$', prop.get('property' '')):
                entity_type = IlluminanceSensor
            elif prop.get('type') == 'binary':
                entity_type = BinarySensor
                sensor_args['value'] = sensor_args['value'] == prop.get(
                    'value_on', True
                )
            elif prop.get('type') == 'enum':
                entity_type = EnumSensor
                sensor_args['values'] = prop.get('values', [])
            elif prop.get('type') == 'numeric':
                entity_type = NumericSensor

            if entity_type:
                sensors.append(
                    cls._to_entity(
                        entity_type, device_info, prop, options=options, **sensor_args
                    )
                )

        return sensors

    @classmethod
    def _get_dimmers(
        cls, device_info: dict, props: dict, options: dict
    ) -> List[Dimmer]:
        """
        A utility method that parses the properties of a device that can be
        mapped to dimmers (or dimmer-like entities).
        """
        return [
            cls._to_entity(
                Dimmer,
                device_info,
                prop,
                options=options,
                value=device_info.get('state', {}).get(prop['property']),
                min=prop.get('value_min'),
                max=prop.get('value_max'),
                unit=prop.get('unit'),
            )
            for prop in [*props.values(), *options.values()]
            if (
                prop.get('property')
                and prop.get('type') == 'numeric'
                and not cls._is_read_only(prop)
            )
        ]

    @classmethod
    def _get_enum_switches(
        cls, device_info: dict, props: dict, options: dict
    ) -> List[EnumSwitch]:
        """
        A utility method that parses the properties of a device that can be
        mapped to switches with enum values.
        """
        return [
            cls._to_entity(
                EnumSwitch,
                device_info,
                prop,
                options=options,
                value=device_info.get('state', {}).get(prop['property']),
                values=prop.get('values', []),
            )
            for prop in [*props.values(), *options.values()]
            if (
                prop.get('access', 0) & 2
                and prop.get('type') == 'enum'
                and prop.get('values')
            )
        ]

    @classmethod
    def _to_entity(  # pylint: disable=redefined-builtin
        cls,
        entity_type: Type[Entity],
        device_info: dict,
        property: dict,
        options: dict,
        **kwargs,
    ) -> Entity:
        """
        Give the information about a device and its properties and options, it
        builds an entity of the right type.
        """
        return entity_type(
            id=f'{device_info["ieee_address"]}:{property["property"]}',
            name=property.get('description', ''),
            is_read_only=cls._is_read_only(property),
            is_write_only=cls._is_write_only(property),
            is_query_disabled=cls._is_query_disabled(property),
            is_configuration=property['property'] in options,
            **kwargs,
        )

    @classmethod
    def _get_light_meta(cls, device_info: dict) -> dict:
        """
        Parse the attributes of a device that can be mapped to lights (or
        light-like entities).
        """
        # pylint: disable=too-many-nested-blocks
        for exposed in (device_info.get('definition', {}) or {}).get('exposes', []):
            if exposed.get('type') == 'light':
                features = exposed.get('features', [])
                switch = {}
                brightness = {}
                temperature = {}
                color = {}

                for feature in features:
                    data = {
                        'is_read_only': cls._is_read_only(feature),
                        'is_write_only': cls._is_write_only(feature),
                        'is_query_disabled': cls._is_query_disabled(feature),
                    }

                    if (
                        feature.get('property') == 'state'
                        and feature.get('type') == 'binary'
                        and 'value_on' in feature
                        and 'value_off' in feature
                    ):
                        switch = {
                            'value_on': feature['value_on'],
                            'value_off': feature['value_off'],
                            'state_name': feature['name'],
                            'value_toggle': feature.get('value_toggle'),
                            **data,
                        }
                    elif (
                        feature.get('property') == 'brightness'
                        and feature.get('type') == 'numeric'
                        and 'value_min' in feature
                        and 'value_max' in feature
                    ):
                        brightness = {
                            'brightness_name': feature['name'],
                            'brightness_min': feature['value_min'],
                            'brightness_max': feature['value_max'],
                            **data,
                        }
                    elif (
                        feature.get('property') == 'color_temp'
                        and feature.get('type') == 'numeric'
                        and 'value_min' in feature
                        and 'value_max' in feature
                    ):
                        temperature = {
                            'temperature_name': feature['name'],
                            'temperature_min': feature['value_min'],
                            'temperature_max': feature['value_max'],
                            **data,
                        }
                    elif (
                        feature.get('property') == 'color'
                        and feature.get('type') == 'composite'
                    ):
                        color_features = feature.get('features', [])
                        for color_feature in color_features:
                            if color_feature.get('property') == 'hue':
                                color.update(
                                    {
                                        'hue_name': color_feature['name'],
                                        'hue_min': color_feature.get('value_min', 0),
                                        'hue_max': color_feature.get(
                                            'value_max', 65535
                                        ),
                                        **data,
                                    }
                                )
                            elif color_feature.get('property') == 'saturation':
                                color.update(
                                    {
                                        'saturation_name': color_feature['name'],
                                        'saturation_min': color_feature.get(
                                            'value_min', 0
                                        ),
                                        'saturation_max': color_feature.get(
                                            'value_max', 255
                                        ),
                                        **data,
                                    }
                                )
                            elif color_feature.get('property') == 'x':
                                color.update(
                                    {
                                        'x_name': color_feature['name'],
                                        'x_min': color_feature.get('value_min', 0.0),
                                        'x_max': color_feature.get('value_max', 1.0),
                                        **data,
                                    }
                                )
                            elif color_feature.get('property') == 'y':
                                color.update(
                                    {
                                        'y_name': color_feature['name'],
                                        'y_min': color_feature.get('value_min', 0),
                                        'y_max': color_feature.get('value_max', 255),
                                        **data,
                                    }
                                )

                return {
                    'friendly_name': device_info.get('friendly_name'),
                    'ieee_address': device_info.get('friendly_name'),
                    **switch,
                    **brightness,
                    **temperature,
                    **color,
                }

        return {}

    @action
    def set_lights(self, *_, lights, **kwargs):
        """
        Set the state for one or more Zigbee lights.
        """
        lights = [lights] if isinstance(lights, str) else lights
        lights = [str(self._ieee_address(light)) for light in lights]
        devices = [self._get_device_info(light) for light in lights]

        for i, dev in enumerate(devices):
            assert dev, f'No such device: {lights[i]}'
            light_meta = self._get_light_meta(dev)
            assert light_meta, f'{dev["name"]} is not a light'
            data = {}

            for attr, value in kwargs.items():
                if attr == 'on':
                    data[light_meta['state_name']] = value
                elif attr in {'brightness', 'bri'}:
                    data[light_meta['brightness_name']] = int(value)
                elif attr in {'temperature', 'ct'}:
                    data[light_meta['temperature_name']] = int(value)
                elif attr in {'saturation', 'sat'}:
                    data['color'] = {
                        **data.get('color', {}),
                        light_meta['saturation_name']: int(value),
                    }
                elif attr == 'hue':
                    data['color'] = {
                        **data.get('color', {}),
                        light_meta['hue_name']: int(value),
                    }
                elif attr == 'xy':
                    data['color'] = {
                        **data.get('color', {}),
                        light_meta['x_name']: float(value[0]),
                        light_meta['y_name']: float(value[1]),
                    }
                elif attr == 'x':
                    data['color'] = {
                        **data.get('color', {}),
                        light_meta['x_name']: float(value),
                    }
                elif attr == 'y':
                    data['color'] = {
                        **data.get('color', {}),
                        light_meta['y_name']: float(value),
                    }
                else:
                    data[attr] = value

            self.device_set(self._preferred_name(dev), values=data)

    def on_mqtt_message(self):
        """
        Overrides :meth:`platypush.plugins.mqtt.MqttPlugin.on_mqtt_message` to
        handle messages from the zigbee2mqtt integration.
        """

        def handler(client: MqttClient, _, msg: mqtt.MQTTMessage):
            topic_idx = len(self.topic_prefix) + 1
            topic = msg.topic[topic_idx:]
            data = msg.payload.decode()
            if not data:
                return

            with contextlib.suppress(ValueError, TypeError):
                data = json.loads(data)

            if topic == 'bridge/state':
                self._process_state_message(client, data)
            elif topic in ['bridge/log', 'bridge/logging']:
                self._process_log_message(client, data)
            elif topic == 'bridge/devices':
                self._process_devices(client, data)
            elif topic == 'bridge/groups':
                self._process_groups(client, data)
            elif isinstance(data, dict):
                name = topic.split('/')[-1]
                if name not in self._info.devices:
                    self.logger.debug('Skipping unknown topic: %s', topic)
                    return

                dev = self._info.devices.get(name)
                assert dev is not None, f'No such device: {name}'
                changed_props = {
                    k: v for k, v in data.items() if v != dev.get('state', {}).get(k)
                }

                if changed_props:
                    self._process_property_update(name, data)
                    self._bus.post(
                        ZigbeeMqttDevicePropertySetEvent(
                            host=client.host,
                            port=client.port,
                            device=name,
                            properties=changed_props,
                        )
                    )

                dev = self._info.devices.get(name)
                if dev:
                    self._info.devices.set_state(
                        dev.get('friendly_name') or dev.get('ieee_address'), data
                    )

        return handler

    def _process_state_message(self, client: MqttClient, msg: str):
        """
        Process a state message.
        """
        if msg == self._info.bridge_state:
            return

        if msg == 'online':
            evt = ZigbeeMqttOnlineEvent
            self._info.bridge_state = BridgeState.ONLINE
        elif msg == 'offline':
            evt = ZigbeeMqttOfflineEvent
            self._info.bridge_state = BridgeState.OFFLINE
            self.logger.warning('The zigbee2mqtt service is offline')
        else:
            return

        self._bus.post(evt(host=client.host, port=client.port))

    # pylint: disable=too-many-branches
    def _process_log_message(self, client, msg):
        """
        Process a log event.
        """

        msg_type = msg.get('type')
        text = msg.get('message')
        args = {'host': client._host, 'port': client._port}

        if msg_type == 'devices':
            devices = {}
            for dev in text or []:
                devices[dev['friendly_name']] = dev
                client.subscribe(self.topic_prefix + '/' + dev['friendly_name'])
        elif msg_type == 'pairing':
            self._bus.post(ZigbeeMqttDevicePairingEvent(device=text, **args))
        elif msg_type in ['device_ban', 'device_banned']:
            self._bus.post(ZigbeeMqttDeviceBannedEvent(device=text, **args))
        elif msg_type in ['device_removed_failed', 'device_force_removed_failed']:
            force = msg_type == 'device_force_removed_failed'
            self._bus.post(
                ZigbeeMqttDeviceRemovedFailedEvent(device=text, force=force, **args)
            )
        elif msg_type == 'device_whitelisted':
            self._bus.post(ZigbeeMqttDeviceWhitelistedEvent(device=text, **args))
        elif msg_type == 'device_renamed':
            self._bus.post(ZigbeeMqttDeviceRenamedEvent(device=text, **args))
        elif msg_type == 'device_bind':
            self._bus.post(ZigbeeMqttDeviceBindEvent(device=text, **args))
        elif msg_type == 'device_unbind':
            self._bus.post(ZigbeeMqttDeviceUnbindEvent(device=text, **args))
        elif msg_type == 'device_group_add':
            self._bus.post(ZigbeeMqttGroupAddedEvent(group=text, **args))
        elif msg_type == 'device_group_add_failed':
            self._bus.post(ZigbeeMqttGroupAddedFailedEvent(group=text, **args))
        elif msg_type == 'device_group_remove':
            self._bus.post(ZigbeeMqttGroupRemovedEvent(group=text, **args))
        elif msg_type == 'device_group_remove_failed':
            self._bus.post(ZigbeeMqttGroupRemovedFailedEvent(group=text, **args))
        elif msg_type == 'device_group_remove_all':
            self._bus.post(ZigbeeMqttGroupRemoveAllEvent(group=text, **args))
        elif msg_type == 'device_group_remove_all_failed':
            self._bus.post(ZigbeeMqttGroupRemoveAllFailedEvent(group=text, **args))
        elif msg_type == 'zigbee_publish_error':
            self.logger.error('zigbee2mqtt error: {}'.format(text))
            self._bus.post(ZigbeeMqttErrorEvent(error=text, **args))
        elif msg.get('level') in ['warning', 'error']:
            log = getattr(self.logger, msg['level'])
            log(
                'zigbee2mqtt {}: {}'.format(
                    msg['level'], text or msg.get('error', msg.get('warning'))
                )
            )

    def _process_devices(self, client: MqttClient, msg):
        """
        Process a list of devices received on the zigbee2mqtt bridge.
        """
        devices_info = {
            device.get('friendly_name', device.get('ieee_address')): device
            for device in msg
        }

        # Subscribe to updates from all the known devices
        event_args = {'host': client.host, 'port': client.port}
        client.subscribe(
            *[self.topic_prefix + '/' + device for device in devices_info.keys()]
        )

        for name, device in devices_info.items():
            # If we haven't cached this device yet, then notify about the
            # connection of a new device.
            if not self._info.devices.get(name):
                self._bus.post(
                    ZigbeeMqttDeviceConnectedEvent(device=name, **event_args)
                )

            # Send a request to fetch all the known properties of this device
            exposes = (device.get('definition', {}) or {}).get('exposes', [])
            payload = self._build_device_get_request(exposes)
            if payload:
                client.publish(
                    self.topic_prefix + '/' + name + '/get',
                    json.dumps(payload),
                )

        # Send a request to fetch all the known properties of this device
        for name in self._info.devices.by_name.copy():
            if name not in devices_info:
                self._bus.post(ZigbeeMqttDeviceRemovedEvent(device=name, **event_args))
                self._info.devices.remove(name)

        for dev in devices_info.values():
            self._info.devices.add(dev)

    def _process_groups(self, client: MqttClient, msg):
        """
        Process an MQTT message containing an updated list of groups.
        """
        event_args = {'host': client.host, 'port': client.port}
        groups_info = {
            group.get('friendly_name', group.get('id')): group for group in msg
        }

        # Trigger ZigbeeMqttGroupAddedEvent for each new group
        for name in groups_info.keys():
            if name not in self._info.groups:
                self._bus.post(ZigbeeMqttGroupAddedEvent(group=name, **event_args))

        # Trigger ZigbeeMqttGroupRemovedEvent for each removed group
        for name in self._info.groups.copy():
            if name not in groups_info:
                self._bus.post(ZigbeeMqttGroupRemovedEvent(group=name, **event_args))
                del self._info.groups[name]

        # Reset the groups cache
        self._info.groups = {group: {} for group in groups_info.keys()}

    def _process_property_update(self, device_name: str, properties: Mapping):
        """
        Process an MQTT message containing a device property update.

        It will appropriately forward an
        :class:`platypush.message.event.entities.EntityUpdateEvent` to the bus.
        """
        device_info = self._info.devices.get(device_name)
        if not (device_info and properties):
            return

        self.publish_entities(
            [
                {
                    **device_info,
                    'state': properties,
                }
            ]
        )


# vim:sw=4:ts=4:et:



from collections import defaultdict
from enum import Enum
from functools import wraps
import re
from typing import Callable, Dict
from uuid import UUID

import bluetooth_numbers
from bleak.uuids import uuid16_dict, uuid128_dict

from platypush.plugins.bluetooth._types import RawServiceClass


def memoized_names(f: Callable[[str], str]):
    """
    Decorator that keeps track of the assigned names to make sure that we don't
    end up with conflicts when assigning service class enum names.
    """
    assigned_names: Dict[str, int] = defaultdict(lambda: 0)

    @wraps(f)
    def wrapper(name: str) -> str:
        name = f(name)
        assigned_names[name] += 1
        transformed_name = (
            f'{name}_{assigned_names[name]}' if assigned_names[name] > 1 else name
        )

        return transformed_name

    return wrapper


@memoized_names
def _service_name_to_enum_name(service_name: str) -> str:
    """
    Convert a service name to an enum-key compatible string.
    """
    ret = service_name.title()
    ret = re.sub(r"\(.+?\)", "", ret)
    ret = re.sub(r"\s+", "_", ret)
    ret = re.sub(r"[^a-zA-Z0-9_]", "", ret)
    ret = re.sub(r"_+", "_", ret)
    return ret.upper()


_service_classes: Dict[RawServiceClass, str] = {
    0x0: "Unknown",
    0x1000: "Service Discovery Server Service Class ID",
    0x1001: "Browse Group Descriptor Service Class ID",
    0x1101: "Serial Port",
    0x1102: "LAN Access Using PPP",
    0x1103: "Dialup Networking",
    0x1104: "IR MC Sync",
    0x1105: "OBEX Object Push",
    0x1106: "OBEX File Transfer",
    0x1107: "IR MC Sync Command",
    0x1108: "Headset",
    0x1109: "Cordless Telephony",
    0x110A: "Audio Source",
    0x110B: "Audio Sink",
    0x110C: "A/V Remote Control Target",
    0x110D: "Advanced Audio Distribution",
    0x110E: "A/V Remote Control",
    0x110F: "A/V Remote Control Controller",
    0x1110: "Intercom",
    0x1111: "Fax",
    0x1112: "Headset Audio Gateway",
    0x1113: "WAP",
    0x1114: "WAP Client",
    0x1115: "PANU",
    0x1116: "NAP",
    0x1117: "GN",
    0x1118: "Direct Printing",
    0x1119: "Reference Printing",
    0x111A: "Basic Imaging Profile",
    0x111B: "Imaging Responder",
    0x111C: "Imaging Automatic Archive",
    0x111D: "Imaging Referenced Objects",
    0x111E: "Handsfree",
    0x111F: "Handsfree Audio Gateway",
    0x1120: "Direct Printing Reference Objects Service",
    0x1121: "Reflected UI",
    0x1122: "Basic Printing",
    0x1123: "Printing Status",
    0x1124: "Human Interface Device Service",
    0x1125: "Hard Copy Cable Replacement",
    0x1126: "HCR Print",
    0x1127: "HCR Scan",
    0x1128: "Common ISDN Access",
    0x112D: "SIM Access",
    0x112E: "Phone Book Access PCE",
    0x112F: "Phone Book Access PSE",
    0x1130: "Phone Book Access",
    0x1131: "Headset NS",
    0x1132: "Message Access Server",
    0x1133: "Message Notification Server",
    0x1134: "Message Notification Profile",
    0x1135: "GNSS",
    0x1136: "GNSS Server",
    0x1137: "3D Display",
    0x1138: "3D Glasses",
    0x1139: "3D Synchronization",
    0x113A: "MPS Profile",
    0x113B: "MPS SC",
    0x113C: "CTN Access Service",
    0x113D: "CTN Notification Service",
    0x113E: "CTN Profile",
    0x1200: "PnP Information",
    0x1201: "Generic Networking",
    0x1202: "Generic File Transfer",
    0x1203: "Generic Audio",
    0x1204: "Generic Telephony",
    0x1205: "UPNP Service",
    0x1206: "UPNP IP Service",
    0x1300: "ESDP UPNP IP PAN",
    0x1301: "ESDP UPNP IP LAP",
    0x1302: "ESDP UPNP L2CAP",
    0x1303: "Video Source",
    0x1304: "Video Sink",
    0x1305: "Video Distribution",
    0x1400: "HDP",
    0x1401: "HDP Source",
    0x1402: "HDP Sink",
}
"""
Directory of known Bluetooth service UUIDs.

See
https://btprodspecificationrefs.blob.core.windows.net/assigned-numbers/Assigned%20Number%20Types/Assigned%20Numbers.pdf,
Section 3.3.
"""

_custom_service_classes: Dict[RawServiceClass, str] = {
    UUID("cba20d00-224d-11e6-9fb8-0002a5d5c51b"): "Switchbot",
}

# Update the base services with the GATT service UUIDs defined in ``bluetooth_numbers``. See
# https://btprodspecificationrefs.blob.core.windows.net/assigned-numbers/Assigned%20Number%20Types/Assigned%20Numbers.pdf,
# Section 3.4
_service_classes.update(bluetooth_numbers.service)

# Extend the service classes with the GATT service UUIDs defined in Bleak
_service_classes.update(_custom_service_classes)
_service_classes.update({UUID(uuid): name for uuid, name in uuid128_dict.items()})
_service_classes.update(uuid16_dict)  # type: ignore

_service_classes_by_name: Dict[str, RawServiceClass] = {
    name: cls for cls, name in _service_classes.items()
}


class _ServiceClassMeta:
    """
    Metaclass for :class:`ServiceClass`.
    """

    value: RawServiceClass
    """ The raw service class value. """

    @classmethod
    def get(cls, value: RawServiceClass) -> "ServiceClass":
        """
        :param value: The raw service class UUID.
        :return: The parsed :class:`ServiceClass` instance, or
            ``ServiceClass.UNKNOWN``.
        """
        try:
            return ServiceClass(value)
        except ValueError:
            try:
                if isinstance(value, UUID):
                    return ServiceClass(int(str(value).upper()[4:8], 16))
            except ValueError:
                pass

            return ServiceClass.UNKNOWN  # type: ignore

    @classmethod
    def by_name(cls, name: str) -> "ServiceClass":
        """
        :param name: The name of the service class.
        :return: The :class:`ServiceClass` instance, or
            ``ServiceClass.UNKNOWN``.
        """
        return (
            ServiceClass(_service_classes_by_name.get(name))
            or ServiceClass.UNKNOWN  # type: ignore
        )

    def __str__(self) -> str:
        return _service_classes.get(
            self.value, ServiceClass.UNKNOWN.value  # type: ignore
        )

    def __repr__(self) -> str:
        return f"<{self.value}: {str(self)}>"


ServiceClass = Enum(  # type: ignore
    "ServiceClass",
    {_service_name_to_enum_name(name): cls for cls, name in _service_classes.items()},
    type=_ServiceClassMeta,
)
""" Enumeration of known Bluetooth services. """



"""
Common Torrents CSV constants.
"""

TORRENT_CSV_API_URL = 'https://torrents-csv.com/service'
""" Default Torrents CSV API base URL. """

TORRENTS_CSV_DOWNLOAD_URL = 'https://git.torrents-csv.com/heretic/torrents-csv-data/raw/branch/main/torrents.csv'
""" Default torrents.csv download URL. """

TORRENTS_CSV_URL_LAST_CHECKED_VAR = '_TORRENTS_CSV_URL_LAST_CHECKED'
""" Environment variable to store the last checked timestamp for the torrents.csv URL. """

TORRENTS_CSV_DEFAULT_CHECK_INTERVAL = 60 * 60 * 24
""" Interval in seconds to re-check the torrents.csv URL (24 hours). """



import re
from typing import Optional, Type


def type_str(t: Optional[Type]) -> Optional[str]:
    """
    :return: A human-readable representation of a type.
    """
    if not t:
        return None

    return re.sub(r"<class '(.*)'>", r'\1', str(t).replace('typing.', ''))



from dataclasses import dataclass
from typing import Optional, Type

from .._serialize import Serializable
from .._utils import type_str


@dataclass
class Argument(Serializable):
    """
    Represents an integration constructor/action parameter.
    """

    name: str
    required: bool = False
    doc: Optional[str] = None
    type: Optional[Type] = None
    default: Optional[str] = None

    def to_dict(self) -> dict:
        return {
            "name": self.name,
            "required": self.required,
            "doc": self.doc,
            "type": type_str(self.type),
            "default": self.default,
        }



import importlib
import inspect
import os
import pkgutil

from platypush.backend import Backend
from platypush.config import Config
from platypush.utils import get_ip_or_hostname

from .logger import logger


def get_http_port():
    from platypush.backend.http import HttpBackend

    http_conf = Config.get('backend.http') or {}
    return http_conf.get('port', HttpBackend.DEFAULT_HTTP_PORT)


def get_routes():
    base_pkg = '.'.join([Backend.__module__, 'http', 'app', 'routes'])
    base_dir = os.path.join(
        os.path.dirname(inspect.getfile(Backend)), 'http', 'app', 'routes'
    )
    routes = []

    for _, mod_name, _ in pkgutil.walk_packages([base_dir], prefix=base_pkg + '.'):
        try:
            module = importlib.import_module(mod_name)
            if hasattr(module, '__routes__'):
                routes.extend(module.__routes__)
        except Exception as e:
            logger().warning('Could not import module %s: %s', mod_name, str(e))
            continue

    return routes


def get_local_base_url():
    http_conf = Config.get('backend.http') or {}
    bind_address = http_conf.get('bind_address')
    if not bind_address or bind_address == '0.0.0.0':
        bind_address = 'localhost'

    return '{proto}://{host}:{port}'.format(
        proto=('https' if http_conf.get('ssl_cert') else 'http'),
        host=bind_address,
        port=get_http_port(),
    )


def get_remote_base_url():
    http_conf = Config.get('backend.http') or {}
    return '{proto}://{host}:{port}'.format(
        proto=('https' if http_conf.get('ssl_cert') else 'http'),
        host=get_ip_or_hostname(),
        port=get_http_port(),
    )



import ssl

from contextlib import contextmanager
from typing import Generator, Iterable, Optional, List, Dict, Union, Any, Tuple

from imapclient import IMAPClient
from imapclient.response_types import Address

from .._model import Mail, TransportEncryption
from .._plugin import MailInPlugin


class MailImapPlugin(MailInPlugin):
    """
    Plugin to interact with a mail server over IMAP.
    """

    @classmethod
    def _matches_url_scheme(cls, scheme: str) -> bool:
        return scheme in ('imap', 'imaps')

    @classmethod
    def default_ports(cls) -> Dict[TransportEncryption, int]:
        return {
            TransportEncryption.NONE: 143,
            TransportEncryption.SSL: 993,
        }

    @contextmanager
    def connect(self) -> Generator[IMAPClient, None, None]:
        has_ssl = self.server.encryption == TransportEncryption.SSL
        context = None
        if has_ssl and self.server.certfile:
            context = ssl.create_default_context()
            context.load_cert_chain(
                certfile=self.server.certfile, keyfile=self.server.keyfile
            )

        client = IMAPClient(
            host=self.server.server,
            port=self.server.port,
            ssl=has_ssl,
            ssl_context=context,
        )

        if self.account.access_token:
            client.oauth2_login(
                self.account.username,
                access_token=self.account.access_token,
                mech=self.account.oauth_mechanism or 'XOAUTH2',
                vendor=self.account.oauth_vendor,
            )
        else:
            pwd = self.account.get_password()
            client.login(self.account.username, pwd)

        yield client
        client.logout()

    @staticmethod
    def _get_folders(data: List[tuple]) -> List[Dict[str, str]]:
        folders = []
        for line in data:
            (flags), delimiter, mailbox_name = line
            folders.append(
                {
                    'name': mailbox_name,
                    'flags': [flag.decode() for flag in flags],
                    'delimiter': delimiter.decode(),
                }
            )

        return folders

    @staticmethod
    def _parse_address(imap_addr: Address) -> Dict[str, Optional[str]]:
        return {
            'name': imap_addr.name.decode() if imap_addr.name else None,
            'route': imap_addr.route.decode() if imap_addr.route else None,
            'email': imap_addr.mailbox.decode() + '@' + imap_addr.host.decode(),
        }

    @classmethod
    def _parse_addresses(
        cls, addresses: Optional[Tuple[Address]] = None
    ) -> Dict[str, Dict[str, str]]:
        ret = {}
        if addresses:
            for addr in addresses:
                addr = cls._parse_address(addr)
                ret[addr['email']] = addr

        return ret

    @classmethod
    def _parse_message(cls, msg_id: int, imap_msg: Dict[bytes, Any]) -> Mail:
        message = {
            'id': msg_id,
            'seq': imap_msg[b'SEQ'],
        }

        if imap_msg.get(b'FLAGS'):
            message['flags'] = [flag.decode() for flag in imap_msg[b'FLAGS'] if flag]
        if b'INTERNALDATE' in imap_msg:
            message['internal_date'] = imap_msg[b'INTERNALDATE']
        if b'RFC822.SIZE' in imap_msg:
            message['size'] = imap_msg[b'RFC822.SIZE']
        if b'ENVELOPE' in imap_msg:
            envelope = imap_msg[b'ENVELOPE']
            message['bcc'] = cls._parse_addresses(envelope.bcc)
            message['cc'] = cls._parse_addresses(envelope.cc)
            message['date'] = envelope.date
            message['from'] = cls._parse_addresses(envelope.from_)
            message['message_id'] = (
                envelope.message_id.decode() if envelope.message_id else None
            )
            message['in_reply_to'] = (
                envelope.in_reply_to.decode() if envelope.in_reply_to else None
            )
            message['reply_to'] = cls._parse_addresses(envelope.reply_to)
            message['sender'] = cls._parse_addresses(envelope.sender)
            message['subject'] = envelope.subject.decode() if envelope.subject else None
            message['to'] = cls._parse_addresses(envelope.to)
        if b'BODY[]' in imap_msg:
            message['content'] = imap_msg[b'BODY[]']

        return Mail(**message)

    @staticmethod
    def _convert_flags(flags: Union[str, Iterable[str]]) -> List[bytes]:
        if isinstance(flags, str):
            flags = [flag.strip() for flag in flags.split(',')]
        return [('\\' + flag).encode() for flag in flags]

    def get_folders(
        self, folder: str = '', pattern: str = '*', **_
    ) -> List[Dict[str, str]]:
        with self.connect() as client:
            data = client.list_folders(directory=folder, pattern=pattern)

        return self._get_folders(data)

    def get_sub_folders(
        self, folder: str = '', pattern: str = '*', **_
    ) -> List[Dict[str, str]]:
        with self.connect() as client:
            data = client.list_sub_folders(directory=folder, pattern=pattern)

        return self._get_folders(data)

    def search(
        self,
        criteria: Union[str, Iterable[str]] = 'ALL',
        folder: str = 'INBOX',
        attributes: Optional[Iterable[str]] = None,
        **_,
    ) -> List[Mail]:
        if not attributes:
            attributes = ['FLAGS', 'INTERNALDATE', 'RFC822.SIZE', 'ENVELOPE']
        else:
            attributes = [attr.upper() for attr in attributes]

        data = {}
        with self.connect() as client:
            client.select_folder(folder, readonly=True)
            ids = client.search(criteria)  # type: ignore
            if ids:
                data = client.fetch(list(ids), attributes)

        return [
            self._parse_message(msg_id, data[msg_id]) for msg_id in sorted(data.keys())
        ]

    def search_unseen_messages(self, folder: str = 'INBOX') -> List[Mail]:
        return self.search(criteria='UNSEEN', directory=folder)

    def search_flagged_messages(self, folder: str = 'INBOX', **_) -> List[Mail]:
        return self.search(criteria='Flagged', directory=folder)

    def search_starred_messages(self, folder: str = 'INBOX', **_) -> List[Mail]:
        return self.search_flagged_messages(folder)

    def sort(
        self,
        folder: str = 'INBOX',
        sort_criteria: Union[str, Iterable[str]] = 'ARRIVAL',
        criteria: Union[str, Iterable[str]] = 'ALL',
    ) -> List[int]:
        with self.connect() as client:
            client.select_folder(folder, readonly=True)
            msg_ids = client.sort(sort_criteria=sort_criteria, criteria=criteria)  # type: ignore
            return msg_ids

    def get_messages(
        self,
        *ids: int,
        folder: str = 'INBOX',
        with_body: bool = True,
        **_,
    ) -> Dict[int, Mail]:
        ret = {}
        with self.connect() as client:
            client.select_folder(folder, readonly=True)
            attrs = ['FLAGS', 'RFC822.SIZE', 'INTERNALDATE', 'ENVELOPE']
            if with_body:
                attrs.append('BODY[]')

            data = client.fetch(ids, attrs)
            for id in ids:  # pylint: disable=redefined-builtin
                msg = data.get(id)
                if not msg:
                    continue

                ret[id] = self._parse_message(id, msg)

        return ret

    def create_folder(self, folder: str, **_):
        with self.connect() as client:
            client.create_folder(folder)

    def rename_folder(self, old_name: str, new_name: str, **_):
        with self.connect() as client:
            client.rename_folder(old_name, new_name)

    def delete_folder(self, folder: str, **_):
        with self.connect() as client:
            client.delete_folder(folder)

    def add_flags(
        self,
        messages: List[int],
        flags: Union[str, Iterable[str]],
        folder: str = 'INBOX',
        **_,
    ):
        with self.connect() as client:
            client.select_folder(folder)
            client.add_flags(messages, self._convert_flags(flags))

    def set_flags(
        self,
        messages: List[int],
        flags: Union[str, Iterable[str]],
        folder: str = 'INBOX',
        **_,
    ):
        with self.connect() as client:
            client.select_folder(folder)
            client.set_flags(messages, self._convert_flags(flags))

    def remove_flags(
        self,
        messages: List[int],
        flags: Union[str, Iterable[str]],
        folder: str = 'INBOX',
        **_,
    ):
        with self.connect() as client:
            client.select_folder(folder)
            client.remove_flags(messages, self._convert_flags(flags))

    def flag_messages(self, messages: List[int], folder: str = 'INBOX', **_):
        return self.add_flags(messages, ['Flagged'], folder=folder)

    def unflag_messages(self, messages: List[int], folder: str = 'INBOX', **_):
        return self.remove_flags(messages, ['Flagged'], folder=folder)

    def delete_messages(
        self,
        messages: List[int],
        folder: str = 'INBOX',
        expunge: bool = True,
        **_,
    ):
        self.add_flags(messages, ['Deleted'], folder=folder)
        if expunge:
            self.expunge_messages(folder=folder, messages=messages)

    def restore_messages(self, messages: List[int], folder: str = 'INBOX', **_):
        return self.remove_flags(messages, ['Deleted'], folder=folder)

    def copy_messages(
        self,
        messages: List[int],
        dest_folder: str,
        source_folder: str = 'INBOX',
        **_,
    ):
        with self.connect() as client:
            client.select_folder(source_folder)
            client.copy(messages, dest_folder)

    def move_messages(
        self,
        messages: List[int],
        dest_folder: str,
        source_folder: str = 'INBOX',
        **_,
    ):
        with self.connect() as client:
            client.select_folder(source_folder)
            client.move(messages, dest_folder)

    def expunge_messages(
        self,
        folder: str = 'INBOX',
        messages: Optional[List[int]] = None,
        **_,
    ):
        with self.connect() as client:
            client.select_folder(folder)
            client.expunge(messages)


# vim:sw=4:ts=4:et:



import json
import logging
from typing import Optional, Union

from sqlalchemy import (
    Boolean,
    Column,
    Float,
    ForeignKey,
    Integer,
    JSON,
    String,
)

from platypush.common.db import is_defined

from .devices import Device

logger = logging.getLogger(__name__)


class Sensor(Device):
    """
    Abstract class for sensor entities. A sensor entity is, by definition, an
    entity with the ``is_read_only`` property set to ``True``.
    """

    __abstract__ = True

    def __init__(self, *args, **kwargs):
        kwargs['is_read_only'] = True
        super().__init__(*args, **kwargs)


if not is_defined('raw_sensor'):

    class RawSensor(Sensor):
        """
        Models a raw sensor, whose value can contain either a string, a
        hex-encoded binary string, or a JSON-encoded object.
        """

        __tablename__ = 'raw_sensor'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        _value = Column(String)
        is_binary = Column(Boolean, default=False)
        """ If ``is_binary`` is ``True``, then ``value`` is a hex string. """
        is_json = Column(Boolean, default=False)
        """
        If ``is_json`` is ``True``, then ``value`` is a JSON-encoded string
        object or array.
        """
        unit = Column(String)

        @property
        def value(self):
            if self._value is None:
                return None
            if self.is_binary and isinstance(self._value, str):
                value = self._value[2:]
                return bytes(
                    [int(value[i : i + 2], 16) for i in range(0, len(value), 2)]
                )
            if self.is_json and isinstance(self._value, (str, bytes)):
                return json.loads(self._value)
            return self._value

        @value.setter
        def value(
            self, value: Optional[Union[str, bytearray, bytes, list, tuple, set, dict]]
        ):
            if isinstance(value, (bytearray, bytes)):
                self._value = '0x' + ''.join([f'{x:02x}' for x in value])
                self.is_binary = True
            elif isinstance(value, (list, tuple, set)):
                self._value = json.dumps(list(value))
                self.is_json = True
            elif isinstance(value, dict):
                self._value = json.dumps(value)
                self.is_json = True
            else:
                self._value = value
                self.is_binary = False
                self.is_json = False

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('numeric_sensor') and not is_defined('percent_sensor'):

    class NumericSensor(Sensor):
        """
        Models a numeric sensor, with a numeric value and an optional min/max
        range.
        """

        __tablename__ = 'numeric_sensor'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        value = Column(Float)
        min = Column(Float)
        max = Column(Float)
        unit = Column(String)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }

    class PercentSensor(NumericSensor):
        """
        A subclass of ``NumericSensor`` that represents a percentage value.
        """

        __tablename__ = 'percent_sensor'

        id = Column(
            Integer, ForeignKey(NumericSensor.id, ondelete='CASCADE'), primary_key=True
        )

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }

        def __init__(self, *args, **kwargs):
            self.min = 0.0
            self.max = 1.0
            self.unit = '%'
            super().__init__(*args, **kwargs)


if not is_defined('binary_sensor'):

    class BinarySensor(Sensor):
        """
        Models a binary sensor, with a binary boolean value.
        """

        __tablename__ = 'binary_sensor'

        def __init__(self, *args, value=None, **kwargs):
            if isinstance(value, str):
                value = value.lower()

            if str(value).lower() in {'1', 't', 'true', 'on'}:
                value = True
            elif str(value).lower() in {'0', 'f', 'false', 'off'}:
                value = False
            elif value is not None:
                logger.warning('Unsupported value for BinarySensor type: %s', value)
                value = None

            super().__init__(*args, value=value, **kwargs)

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        value = Column(Boolean)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('enum_sensor'):

    class EnumSensor(Sensor):
        """
        Models an enum sensor, whose value belongs to a set of pre-defined values.
        """

        __tablename__ = 'enum_sensor'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        value = Column(String)
        values = Column(JSON)
        """ Possible values for the sensor, as a JSON array. """

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }


if not is_defined('composite_sensor'):

    class CompositeSensor(Sensor):
        """
        A composite sensor is a sensor whose value can be mapped to a JSON
        object (either a dictionary or an array)
        """

        __tablename__ = 'composite_sensor'

        id = Column(
            Integer, ForeignKey(Device.id, ondelete='CASCADE'), primary_key=True
        )
        value = Column(JSON)

        __table_args__ = {'extend_existing': True}
        __mapper_args__ = {
            'polymorphic_identity': __tablename__,
        }



from platypush.commands import Command


class StopCommand(Command):
    """
    Stop the application.
    """

    def __call__(self, app, *_, **__):
        self.logger.info('Received StopApplication command.')
        app.stop()


class RestartCommand(Command):
    """
    Restart the application.
    """

    def __call__(self, app, *_, **__):
        self.logger.info('Received RestartApplication command.')
        app.restart()



import threading
from typing import Callable, Optional, Type


def OrEvent(*events, cls: Type = threading.Event):
    """
    Wrapper for threading.Event that allows to create an event that is
    set if any of the given events are set.

    Adapted from
    https://stackoverflow.com/questions/12317940/python-threading-can-i-sleep-on-two-threading-events-simultaneously#12320352.

    :param events: The events to be checked.
    :param cls: The class to be used for the event. Default: threading.Event.
    """

    or_event = cls()

    def changed():
        bools = [e.is_set() for e in events]
        if any(bools):
            or_event.set()
        else:
            or_event.clear()

    def _to_or(e, changed_callback: Callable[[], None]):
        if not hasattr(e, "_set"):
            e._set = e.set
        if not hasattr(e, "_clear"):
            e._clear = e.clear

        e.changed = changed_callback
        e.set = lambda: _or_set(e)
        e.clear = lambda: _clear_or(e)

    def _clear_or(e):
        e._clear()
        e.changed()

    def _or_set(e):
        try:
            e._set()
            e.changed()
        except RecursionError:
            pass

    for e in events:
        _to_or(e, changed)

    changed()
    return or_event


def wait_for_either(
    *events, timeout: Optional[float] = None, cls: Type = threading.Event
):
    """
    Wait for any of the given events to be set.

    :param events: The events to be checked.
    :param timeout: The maximum time to wait for the event to be set. Default: None.
    :param cls: The class to be used for the event. Default: threading.Event.
    """
    return OrEvent(*events, cls=cls).wait(timeout=timeout)



import json
import multiprocessing

from platypush.backend.http.app.utils import logger
from platypush.backend.http.media.handlers import MediaHandler
from platypush.message import Message
from platypush.utils import get_redis

from ._constants import MEDIA_MAP_VAR, MediaMap

media_map_lock = multiprocessing.RLock()


def load_media_map() -> MediaMap:
    """
    Load the media map from the server.
    """
    with media_map_lock:
        redis = get_redis()
        try:
            media_map = json.loads(
                ((redis.mget(MEDIA_MAP_VAR) or [None])[0] or b'{}').decode()  # type: ignore
            )
        except Exception as e:
            logger().warning('Could not load media map: %s', e)
            return {}

    return {
        media_id: MediaHandler.build(**media_info)
        for media_id, media_info in media_map.items()
    }


def save_media_map(new_map: MediaMap):
    """
    Updates the stored media map on the server.
    """
    with media_map_lock:
        redis = get_redis()
        redis.mset({MEDIA_MAP_VAR: json.dumps(new_map, cls=Message.Encoder)})



import time
import re
from threading import RLock, Timer
from typing import Optional, Sequence, Union

import rtmidi

from platypush.message.event.midi import MidiMessageEvent
from platypush.plugins import RunnablePlugin, action

from ._model import MidiDevice, MidiDeviceType, MidiDevices


class MidiPlugin(RunnablePlugin):
    """
    Virtual MIDI controller plugin. It allows you to send custom MIDI messages
    to any connected MIDI devices and listen for MIDI events.
    """

    _played_notes = set()

    def __init__(
        self,
        device_name: str = 'Platypush MIDI plugin',
        midi_devices: Optional[Sequence[Union[str, int]]] = None,
        poll_interval: Optional[float] = 5.0,
        event_resolution: Optional[float] = None,
        **kwargs,
    ):
        """
        :param device_name: Name of the MIDI device associated to this plugin.
        :param midi_devices: List of MIDI devices to open and monitor for
            events, by name or by port number. If set, and ``poll_interval``
            is set, then the plugin will only listen for events from these
            devices. If not set, and ``poll_interval`` is set, then the plugin
            will listen for events from all available MIDI devices (default).
        :param poll_interval: How often the plugin should scan for new MIDI
            devices. Set this to 0 or null to disable polling.
        :param event_resolution: If set, then the plugin will throttle
            MIDI events to the specified time resolution. If an event is
            triggered within the specified time resolution, then the event will
            be throttled and the last event will be discarded. This is useful
            to avoid sending too many MIDI messages in a short time frame.
            Default: no throttling.
        """
        super().__init__(poll_interval=poll_interval, **kwargs)

        self._devices = MidiDevices()
        self.device_name = device_name
        self._devices_to_monitor = {
            self._to_device_id(dev) for dev in (midi_devices or [])
        }
        self._midi_out = None
        self._midi_in = None
        self._last_midi_event = None
        self._last_event_trigger_time = None
        self._event_timer = None
        self._event_timer_lock = RLock()
        self._event_resolution = event_resolution
        self.logger.info(
            'Initialized MIDI plugin on virtual device %s', self.device_name
        )

    @staticmethod
    def _to_device_id(device: Union[str, int]) -> Union[str, int]:
        try:
            device = int(device)
        except ValueError:
            pass

        return device

    @property
    def midi_in(self) -> rtmidi.MidiIn:  # type: ignore
        if not self._midi_in:
            self._midi_in = rtmidi.MidiIn()  # type: ignore
            self._midi_in.open_port()

        return self._midi_in

    @property
    def midi_out(self) -> rtmidi.MidiOut:  # type: ignore
        if not self._midi_out:
            self._midi_out = rtmidi.MidiOut()  # type: ignore
            self._midi_out.open_port()

        return self._midi_out

    def start_event_timer(self):
        def flush_last_event():
            self._last_event_trigger_time = time.time()

            if not self._last_midi_event:
                return

            self._bus.post(self._last_midi_event)

        with self._event_timer_lock:
            if not self._event_timer:
                self._event_timer = Timer(self._event_resolution or 0, flush_last_event)

                self._event_timer.start()

            return self._event_timer

    def _on_midi_message(self, device: Optional[MidiDevice] = None):
        def callback(message, data):
            # rtmidi will provide a tuple in the format
            # (midi_message, time_since_last_event)
            # delay = message[1]
            message = message[0]
            self._last_midi_event = event = MidiMessageEvent(
                message=message,
                data=data,
                device=device.name if device else None,
                port=device.port if device else None,
            )

            if self._event_resolution and self._last_event_trigger_time:
                event_delta = time.time() - self._last_event_trigger_time
                if event_delta < self._event_resolution:
                    self.start_event_timer()
                    self.logger.debug('Skipping throttled message: %s', message)
                    return

            self._last_event_trigger_time = time.time()
            self._bus.post(event)

        return callback

    @action
    def send_message(
        self, values: Sequence[int], device: Optional[Union[int, str]] = None
    ):
        """
        :param values: Values is expected to be a list containing the MIDI
            command code and the command parameters - `see reference
            <https://ccrma.stanford.edu/~craig/articles/linuxmidi/misc/essenmidi.html>`_.

            Available MIDI commands:

                 * ``0x80``    Note Off
                 * ``0x90``    Note On
                 * ``0xA0``    Aftertouch
                 * ``0xB0``    Continuous controller
                 * ``0xC0``    Patch change
                 * ``0xD0``    Channel Pressure
                 * ``0xE0``    Pitch bend
                 * ``0xF0``    Start of system exclusive message
                 * ``0xF1``    MIDI Time Code Quarter Frame (Sys Common)
                 * ``0xF2``    Song Position Pointer (Sys Common)
                 * ``0xF3``    Song Select
                 * ``0xF6``    Tune Request (Sys Common)
                 * ``0xF7``    End of system exclusive message
                 * ``0xF8``    Timing Clock (Sys Realtime)
                 * ``0xFA``    Start (Sys Realtime)
                 * ``0xFB``    Continue (Sys Realtime)
                 * ``0xFC``    Stop (Sys Realtime)
                 * ``0xFE``    Active Sensing (Sys Realtime)
                 * ``0xFF``    System Reset (Sys Realtime)

        :param device: MIDI port to send the message to, by number or by name.
            If None then the message will be sent to the default port allocated
            for the plugin.
        """
        if isinstance(device, (str, int)):
            if isinstance(device, str):
                dev = self._devices.by_type(MidiDeviceType.OUTPUT).get(device)
            else:
                dev = self._devices.by_port(device)

            assert dev, f'Could not find device by name {device}'
            assert (
                dev.device_type == MidiDeviceType.OUTPUT
            ), f'The device {device} is not an output device'

            midi_out = dev.midi_out
            if not (midi_out and midi_out.is_port_open()):
                dev.open()
        else:
            midi_out = self.midi_out

        assert midi_out and midi_out.is_port_open(), 'No MIDI output port available'
        midi_out.send_message(values)

    @action
    def play(self, note: int, velocity: int, duration: float = 0):
        """
        Play a note with selected velocity and duration.

        :param note: MIDI note in range 0-127 with #60 = C4
        :param velocity: MIDI note velocity in range 0-127
        :param duration: Note duration in seconds. Pass 0 if you don't want the note to get off
        """
        self.send_message([0x90, note, velocity])  # Note on
        self._played_notes.add(note)

        if duration:
            time.sleep(duration)
            self.send_message([0x80, note, 0])  # Note off
            self._played_notes.remove(note)

    @action
    def release(self, note: int):
        """
        Release a played note.

        :param note: MIDI note in range 0-127 with #60 = C4
        """
        self.send_message([0x80, note, 0])  # Note off
        self._played_notes.remove(note)

    @action
    def release_all(self):
        """
        Release all the notes being played.
        """
        played_notes = self._played_notes.copy()
        for note in played_notes:
            self.release(note)

    @action
    def query(self):
        """
        :returns: dict: A list of the available MIDI ports with index and name.
            Format: ``port_index: device_name``.

              .. code-block:: json

                {
                  "in": {
                    0: "Midi Through:Midi Through Port-0 14:0",
                    1: "MPK mini 3:MPK mini 3 MIDI 1 32:0",
                    2: "X-TOUCH MINI:X-TOUCH MINI MIDI 1 36:0",
                    3: "RtMidiOut Client:Platypush MIDI plugin 129:0"
                  },
                  "out": {
                    0: "Midi Through:Midi Through Port-0 14:0",
                    1: "MPK mini 3:MPK mini 3 MIDI 1 32:0",
                    2: "X-TOUCH MINI:X-TOUCH MINI MIDI 1 36:0"
                  }
                }

        """
        in_ports = {int(port): dev for port, dev in enumerate(self.midi_in.get_ports())}
        out_ports = {
            int(port): dev for port, dev in enumerate(self.midi_out.get_ports())
        }

        for device_type, devices in (
            (MidiDeviceType.INPUT, in_ports),
            (MidiDeviceType.OUTPUT, out_ports),
        ):
            for port, name in devices.items():
                dev = MidiDevice(
                    name=name,
                    port=port,
                    device_type=device_type,
                    callback=self._on_midi_message,
                )

                self._devices.add(dev)
                dev = self._devices.by_port(port)
                assert dev, f'Could not find device by port {port}'

                if not self._is_self(dev) and (
                    not self._devices_to_monitor
                    or port in self._devices_to_monitor
                    or name in self._devices_to_monitor
                ):
                    dev.open()

        return {
            'in': in_ports,
            'out': out_ports,
        }

    def _is_self(self, device: MidiDevice):
        return device.name == self.device_name or re.search(
            r'^RtMidi(In|Out) Client:', device.name
        )

    def cleanup(self):
        self.release_all()
        if self._midi_in:
            self._midi_in.close_port()
            self._midi_in = None

        if self._midi_out:
            self._midi_out.close_port()
            self._midi_out = None

        self._devices.close()

    def main(self):
        # Don't run the polling logic if the poll_interval is not set
        if not self.poll_interval:
            self.wait_stop()
            return

        while not self.should_stop():
            self.query()
            self.wait_stop(self.poll_interval)

    def stop(self):
        self.cleanup()
        super().stop()


# vim:sw=4:ts=4:et:



import datetime
import os
import re
import threading
import time

from sqlalchemy import create_engine
from sqlalchemy.sql.expression import func

from platypush.config import Config
from platypush.plugins.media import MediaPlugin
from platypush.plugins.media.search import MediaSearcher

from .db import (
    Base,
    MediaDirectory,
    MediaFile,
    MediaFileToken,
    MediaToken,
    Session,
)

from .metadata import get_metadata

_db_lock = threading.RLock()


class LocalMediaSearcher(MediaSearcher):
    """
    This class will search for media in the local configured directories. It
    will index the media files for a faster search, it will detect which
    directories have been changed since the last scan and re-index their content
    if needed.
    """

    _filename_separators = r'[.,_\-@()\[\]\{\}\s\'\"]+'

    def __init__(self, dirs, *args, **kwargs):
        super().__init__(*args, **kwargs)
        self.dirs = dirs
        db_dir = os.path.join(Config.get_workdir(), 'media')
        os.makedirs(db_dir, exist_ok=True)
        self.db_file = os.path.join(db_dir, 'media.db')
        self._db_engine = None

    def _get_db_session(self):
        if not self._db_engine:
            self._db_engine = create_engine(
                'sqlite:///{}'.format(self.db_file),
                connect_args={'check_same_thread': False},
            )

            Base.metadata.create_all(self._db_engine)
            Session.configure(bind=self._db_engine)

        return Session()

    @staticmethod
    def _get_or_create_dir_entry(session, path):
        record = session.query(MediaDirectory).filter_by(path=path).first()
        if record is None:
            record = MediaDirectory.build(path=path)
            session.add(record)

        session.commit()
        return record

    @classmethod
    def _get_last_modify_time(cls, path, recursive=False):
        return (
            max([os.path.getmtime(p) for p, _, _ in os.walk(path)])
            if recursive
            else os.path.getmtime(path)
        )

    @classmethod
    def _has_directory_changed_since_last_indexing(cls, dir_record):
        if not dir_record.last_indexed_at:
            return True

        return (
            datetime.datetime.fromtimestamp(cls._get_last_modify_time(dir_record.path))
            > dir_record.last_indexed_at
        )

    @classmethod
    def _matches_query(cls, filename, query):
        filename = filename.lower()
        query_tokens = [
            _.lower() for _ in re.split(cls._filename_separators, query.strip())
        ]

        return all(token in filename for token in query_tokens)

    @classmethod
    def _sync_token_records(cls, session, *tokens):
        tokens = list(tokens)
        if not tokens:
            return []

        records = {
            record.token: record
            for record in session.query(MediaToken)
            .filter(MediaToken.token.in_(tokens))
            .all()
        }

        for token in tokens:
            if token in records:
                continue
            record = MediaToken.build(token=token)
            session.add(record)
            records[token] = record

        session.commit()
        return session.query(MediaToken).filter(MediaToken.token.in_(tokens)).all()

    @classmethod
    def _get_file_records(cls, dir_record, session):
        return session.query(MediaFile).filter_by(directory_id=dir_record.id).all()

    def scan(self, media_dir, session=None, dir_record=None):
        """
        Scans a media directory and stores the search results in the internal
        SQLite index
        """

        if not session:
            session = self._get_db_session()

        self.logger.info('Indexing directory {}'.format(media_dir))
        index_start_time = time.time()

        if not dir_record:
            dir_record = self._get_or_create_dir_entry(session, media_dir)

        if not os.path.isdir(media_dir):
            self.logger.info(
                'Directory {} is no longer accessible, removing it'.format(media_dir)
            )
            session.query(MediaDirectory).filter(
                MediaDirectory.path == media_dir
            ).delete(synchronize_session='fetch')
            return

        media_files = []
        stored_file_records = {
            f.path: f for f in self._get_file_records(dir_record, session)
        }

        for path, _, files in os.walk(media_dir):
            for filename in files:
                filepath = os.path.join(path, filename)

                if filepath in stored_file_records:
                    # stored_file_records will be used to keep track of the
                    # files that have been removed from media_dir. If the file
                    # is still there, don't index it again
                    del stored_file_records[filepath]
                    continue

                if not MediaPlugin.is_video_file(
                    filename
                ) and not MediaPlugin.is_audio_file(filename):
                    continue

                self.logger.info('Scanning item %s', filepath)
                tokens = [
                    _.lower()
                    for _ in re.split(self._filename_separators, filename.strip())
                ]

                token_records = self._sync_token_records(session, *tokens)
                file_record = MediaFile.build(directory_id=dir_record.id, path=filepath)
                session.add(file_record)
                session.commit()

                file_record = (
                    session.query(MediaFile)
                    .filter_by(directory_id=dir_record.id, path=filepath)
                    .one()
                )

                for token_record in token_records:
                    file_token = MediaFileToken.build(
                        file_id=file_record.id, token_id=token_record.id
                    )
                    session.add(file_token)

                media_files.append(file_record)

        # stored_file_records should now only contain the records of the files
        # that have been removed from the directory
        if stored_file_records:
            self.logger.info(
                'Removing references to {} deleted media items from {}'.format(
                    len(stored_file_records), media_dir
                )
            )

            session.query(MediaFile).filter(
                MediaFile.id.in_([record.id for record in stored_file_records.values()])
            ).delete(synchronize_session='fetch')

        dir_record.last_indexed_at = datetime.datetime.now()  # type: ignore
        self.logger.info(
            'Scanned {} in {} seconds'.format(
                media_dir, int(time.time() - index_start_time)
            )
        )

        session.commit()

        # Start the metadata scan in a separate thread
        threading.Thread(
            target=self._metadata_scan_thread, args=(media_files,), daemon=True
        ).start()

    def _metadata_scan_thread(self, records):
        """
        Thread that will scan the media files in the given paths and update
        their metadata.
        """
        paths = [record.path for record in records]
        metadata = get_metadata(*paths)
        session = self._get_db_session()

        for record, data in zip(records, metadata):
            record = session.merge(record)
            record.duration = data.get('duration')  # type: ignore
            record.width = data.get('width')  # type: ignore
            record.height = data.get('height')  # type: ignore
            record.image = data.get('image')  # type: ignore
            record.created_at = data.get('created_at')  # type: ignore
            session.add(record)

        session.commit()

    def search(self, query, **_):
        """
        Searches in the configured media directories given a query. It uses the
        built-in SQLite index if available. If any directory has changed since
        the last scan then it will be indexed again and the up-to-date results
        will be returned.
        """

        session = self._get_db_session()
        results = {}

        with _db_lock:
            for media_dir in self.dirs:
                self.logger.info('Searching {} for "{}"'.format(media_dir, query))
                dir_record = self._get_or_create_dir_entry(session, media_dir)

                if self._has_directory_changed_since_last_indexing(dir_record):
                    self.logger.info(
                        '{} has changed since last indexing, '.format(media_dir)
                        + 're-indexing'
                    )

                    self.scan(media_dir, session=session, dir_record=dir_record)

                query_tokens = [
                    _.lower()
                    for _ in re.split(self._filename_separators, query.strip())
                ]

                for file_record in session.query(MediaFile).where(
                    MediaFile.id.in_(
                        session.query(MediaFile.id)
                        .join(MediaFileToken)
                        .join(MediaToken)
                        .filter(MediaToken.token.in_(query_tokens))
                        .group_by(MediaFile.id)
                        .having(
                            func.count(MediaFileToken.token_id) >= len(query_tokens)
                        )
                    )
                ):
                    if os.path.isfile(file_record.path):
                        results[file_record.path] = {
                            'url': 'file://' + file_record.path,
                            'title': os.path.basename(file_record.path),
                            'size': os.path.getsize(file_record.path),
                            'duration': file_record.duration,
                            'width': file_record.width,
                            'height': file_record.height,
                            'image': file_record.image,
                            'created_at': file_record.created_at,
                        }

        return results.values()



import datetime
import enum
import json
from typing import List, Dict, Optional

import requests

from twilio.rest import Client
from twilio.base import values

from platypush.message import Message
from platypush.plugins import Plugin, action


class TwilioPhoneNumberType(enum.Enum):
    MOBILE = 'mobile'
    LOCAL = 'local'
    TOLL_FREE = 'toll_free'


class TwilioPlugin(Plugin):
    """
    The Twilio plugin allows you to send messages and WhatsApp texts and make
    programmable phone call by using a Twilio account. Note that some features
    may require a Premium account.
    """

    _api_base_url = 'https://api.twilio.com'

    def __init__(
        self,
        account_sid: str,
        auth_token: str,
        address_sid: Optional[str] = None,
        phone_number: Optional[str] = None,
        address_book: Optional[Dict[str, str]] = None,
        **kwargs
    ):
        """
        :param account_sid: Account SID.
        :param auth_token: Account authentication token.
        :param address_sid: SID of the default physical address - required to
            register a new number in some countries.
        :param phone_number: Default phone number associated to the account to
            be used for messages and calls.
        :param address_book: ``name``->``phone_number`` mapping of contacts.
            You can use directly these names to send messages and make calls
            instead of the full phone number.
        """
        super().__init__(**kwargs)
        self.account_sid = account_sid
        self.auth_token = auth_token
        self.address_sid = address_sid
        self.phone_number = phone_number
        self.address_book = address_book or {}
        self.client = Client(account_sid, auth_token)

    @action
    def get_available_phone_numbers(self, country: str, number_type: str) -> List[dict]:
        """
        Get a list of phone numbers of a certain type available for a certain country.

        :param country: Country code (e.g. ``US`` or ``NL``).
        :param number_type: Phone number type - e.g. ``mobile``, ``local`` or
            ``toll_free``.
        :return: A list of the available phone numbers with their properties
            and capabilities. Example:

                .. code-block:: json

                    [
                      {
                        "friendly_name": "+311234567890",
                        "phone_number": "+311234567890",
                        "lata": null,
                        "rate_center": null,
                        "latitude": null,
                        "longitude": null,
                        "locality": null,
                        "region": null,
                        "postal_code": null,
                        "iso_country": "NL",
                        "address_requirements": "any",
                        "beta": false,
                        "capabilities": {
                          "voice": true,
                          "SMS": true,
                          "MMS": false,
                          "fax": false
                        }
                      }
                  ]

        """
        phone_numbers = self.client.available_phone_numbers(country.upper()).fetch()
        resp = requests.get(
            self._api_base_url + phone_numbers.uri,
            auth=(self.account_sid, self.auth_token),
        ).json()
        assert (
            'subresource_uris' in resp
        ), 'No available phone numbers found for the country {}'.format(country)
        assert (
            number_type in resp['subresource_uris']
        ), 'No "{}" phone numbers available - available types: {}'.format(
            number_type, list(resp['subresource_uris'].keys())
        )

        resp = requests.get(
            self._api_base_url + resp['subresource_uris'][number_type],
            auth=(self.account_sid, self.auth_token),
        ).json()

        phone_numbers = resp['available_phone_numbers']
        assert len(phone_numbers), 'No phone numbers available'
        return phone_numbers

    @action
    def create_address(
        self,
        customer_name: str,
        street: str,
        city: str,
        region: str,
        postal_code: str,
        iso_country: str,
    ):
        """
        Create a new address associated to your account.

        :param customer_name: Full name of the customer.
        :param street: Street name.
        :param city: City name.
        :param region: Region name.
        :param postal_code: Postal code.
        :param iso_country: ISO code of the country.
        :return: Details of the newly created address. Example:

            .. code-block:: json

                {
                  "account_sid": "ACXXX",
                  "city": "city",
                  "customer_name": "customer_name",
                  "date_created": "Tue, 18 Aug 2015 17:07:30 +0000",
                  "date_updated": "Tue, 18 Aug 2015 17:07:30 +0000",
                  "emergency_enabled": false,
                  "friendly_name": null,
                  "iso_country": "US",
                  "postal_code": "postal_code",
                  "region": "region",
                  "sid": "ADXXX",
                  "street": "street",
                  "validated": false,
                  "verified": false,
                  "uri": "/2010-04-01/Accounts/ACXXX/Addresses/ADXXX.json"
                }

        """
        address = self.client.addresses.create(
            customer_name=customer_name,
            street=street,
            city=city,
            region=region,
            postal_code=postal_code,
            iso_country=iso_country,
        )

        return address._properties

    @action
    def register_phone_number(
        self,
        phone_number: str,
        friendly_name: Optional[str] = None,
        address_sid: Optional[str] = None,
        sms_url: Optional[str] = None,
        sms_fallback_url: Optional[str] = None,
        status_callback: Optional[str] = None,
        voice_caller_id_lookup: bool = True,
        voice_url: Optional[str] = None,
        voice_fallback_url: Optional[str] = None,
        area_code: Optional[str] = None,
    ) -> dict:
        """
        Request to allocate a phone number on your Twilio account. The phone
        number should first be displayed as available in
        :meth:`.get_available_phone_numbers`.

        :param phone_number: Phone number to be allocated.
        :param friendly_name: A string used to identify your new phone number.
        :param address_sid: Address SID. NOTE: some countries may require you
            to specify a valid address in order to register a new phone number
            (see meth:`create_address`). If none is specified then the
            configured ``address_sid`` (if available) will be applied.
        :param sms_url: URL to call when an SMS is received.
        :param sms_fallback_url: URL to call when an error occurs on SMS
            delivery/receipt.
        :param status_callback: URL to call when a status change occurs.
        :param voice_caller_id_lookup: Whether to perform ID lookup for
            incoming caller numbers.
        :param voice_url: URL to call when the number receives a call.
        :param voice_fallback_url: URL to call when a call fails.
        :param area_code: Override the area code for the new number.
        :return: Status of the newly created number. Example:

            .. code-block:: json

                {
                  "account_sid": "ACXXX",
                  "address_requirements": "none",
                  "address_sid": "ADXXX",
                  "api_version": "2010-04-01",
                  "beta": false,
                  "capabilities": {
                    "voice": true,
                    "sms": false,
                    "mms": true,
                    "fax": false
                  },
                  "date_created": "Thu, 30 Jul 2015 23:19:04 +0000",
                  "date_updated": "Thu, 30 Jul 2015 23:19:04 +0000",
                  "emergency_status": "Active",
                  "emergency_address_sid": "ADXXX",
                  "friendly_name": "friendly_name",
                  "identity_sid": "RIXXX",
                  "origin": "origin",
                  "phone_number": "+18089255327",
                  "sid": "PNXXX",
                  "sms_application_sid": "APXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "sms_fallback_method": "GET",
                  "sms_fallback_url": "https://example.com",
                  "sms_method": "GET",
                  "sms_url": "https://example.com",
                  "status_callback": "https://example.com",
                  "status_callback_method": "GET",
                  "trunk_sid": null,
                  "uri": "/2010-04-01/Accounts/ACXXX/IncomingPhoneNumbers/PNXXX.json",
                  "voice_application_sid": "APXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "voice_caller_id_lookup": false,
                  "voice_fallback_method": "GET",
                  "voice_fallback_url": "https://example.com",
                  "voice_method": "GET",
                  "voice_url": "https://example.com",
                  "bundle_sid": "BUXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXXX",
                  "voice_receive_mode": "voice",
                  "status": "in-use"
                }

        """
        status = self.client.incoming_phone_numbers.create(
            phone_number=phone_number,
            friendly_name=friendly_name,
            sms_url=sms_url,
            sms_fallback_url=sms_fallback_url,
            status_callback=status_callback,
            voice_caller_id_lookup=voice_caller_id_lookup,
            voice_url=voice_url,
            voice_fallback_url=voice_fallback_url,
            area_code=area_code,
            address_sid=address_sid or self.address_sid,
        )

        return status._properties

    @action
    def send_message(
        self,
        body: str,
        to: str,
        from_: Optional[str] = None,
        status_callback: Optional[str] = None,
        max_price: Optional[str] = None,
        attempt: Optional[int] = None,
        validity_period: Optional[int] = None,
        smart_encoded: bool = True,
        media_url: Optional[str] = None,
    ) -> dict:
        """
        Send an SMS/MMS.

        Note: WhatsApp messages are also supported (and free of charge),
        although the functionality is currently quite limited. Full support is
        only available to WhatsApp Business profiles and indipendent software
        vendors approved by WhatsApp. If that's not the case, you can send
        WhatsApp messages through the Twilio Test account/number - as of now
        the ``from_`` field should be ``whatsapp:+14155238886`` and the ``to``
        field should be ``whatsapp:+<phone_number_here>``. More information
        `here <https://www.twilio.com/docs/whatsapp>`_.

        :param body: Message body.
        :param to: Recipient number or address book name.
        :param from_: Sender number. If none is specified then the default
            configured ``phone_number`` will be used if available.
        :param status_callback: The URL to call to send status information to
            the application.
        :param max_price: The total maximum price up to 4 decimal places in US
            dollars acceptable for the message to be delivered.
        :param attempt: Total numer of attempts made , this inclusive to send
            out the message.
        :param validity_period: The number of seconds that the message can
            remain in our outgoing queue.
        :param smart_encoded: Whether to detect Unicode characters that have a
            similar GSM-7 character and replace them.
        :param media_url: The URL of the media to send with the message.
        :return: A mapping representing the status of the delivery. Example:

            .. code-block:: json

                {
                  "account_sid": "ACXXX",
                  "api_version": "2010-04-01",
                  "body": "Sent from your Twilio trial account - It works!",
                  "date_created": "2020-08-17T16:32:09.341",
                  "date_updated": "2020-08-17T16:32:09.526",
                  "date_sent": null,
                  "direction": "outbound-api",
                  "error_code": null,
                  "error_message": null,
                  "from_": "+XXXXXXXXX",
                  "messaging_service_sid": null,
                  "num_media": "0",
                  "num_segments": "1",
                  "price": null,
                  "price_unit": "USD",
                  "sid": "XXXXXXXXXXXXX",
                  "status": "queued",
                  "subresource_uris": {
                    "media": "/2010-04-01/Accounts/ACXXX/Messages/SMXXX/Media.json"
                  },
                  "to": "+XXXXXXXXXXXXXX",
                  "uri": "/2010-04-01/Accounts/ACXXX/Messages/SMXXX.json"
                }

        """
        assert (
            from_ or self.phone_number
        ), 'No valid sender phone number specified nor configured'

        to = self.address_book.get(to, to)
        status = self.client.messages.create(
            body=body,
            from_=from_ or self.phone_number,
            to=to,
            media_url=media_url or values.unset,
            max_price=max_price or values.unset,
            attempt=str(attempt) if attempt else values.unset,
            status_callback=status_callback or values.unset,
            validity_period=str(validity_period) if validity_period else values.unset,
            smart_encoded=smart_encoded,
        )

        return status._properties

    @action
    def list_messages(
        self,
        to: Optional[str] = None,
        from_: Optional[str] = None,
        date_sent_before: Optional[str] = None,
        date_sent: Optional[str] = None,
        date_sent_after: Optional[str] = None,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[dict]:
        """
        List all messages matching the specified criteria.

        :param to: Recipient phone number or address book name.
        :param from_: Sender phone number.
        :param date_sent_before: Maximum date filter (ISO format: YYYYMMDD with or without time).
        :param date_sent: Date filter (ISO format: YYYYMMDD with or without time).
        :param date_sent_after: Minimum date filter (ISO format: YYYYMMDD with or without time).
        :param limit: Maximum number of messages to be returned.
        :param page_size: Maximum number of messages per page.
        :return: List of selected messages. Example:

            .. code-block:: json

                {
                  "account_sid": "ACXXX",
                  "api_version": "2010-04-01",
                  "body": "testing",
                  "date_created": "Fri, 24 May 2019 17:44:46 +0000",
                  "date_sent": "Fri, 24 May 2019 17:44:50 +0000",
                  "date_updated": "Fri, 24 May 2019 17:44:50 +0000",
                  "direction": "outbound-api",
                  "error_code": null,
                  "error_message": null,
                  "from": "+12019235161",
                  "messaging_service_sid": null,
                  "num_media": "0",
                  "num_segments": "1",
                  "price": "-0.00750",
                  "price_unit": "USD",
                  "sid": "SMded05904ccb347238880ca9264e8fe1c",
                  "status": "sent",
                  "subresource_uris": {
                    "media": "/2010-04-01/Accounts/ACXXX/Messages/SMXXX/Media.json",
                    "feedback": "/2010-04-01/Accounts/ACXXX/Messages/SMXXX/Feedback.json"
                  },
                  "to": "+18182008801",
                  "uri": "/2010-04-01/Accounts/ACXXX/Messages/SMXXX.json"
                }

        """
        if to:
            to = self.address_book.get(to, to)

        messages = self.client.messages.list(
            to=to,
            from_=from_,
            date_sent_before=datetime.datetime.fromisoformat(date_sent_before)
            if date_sent_before
            else None,
            date_sent_after=datetime.datetime.fromisoformat(date_sent_after)
            if date_sent_after
            else None,
            date_sent=datetime.datetime.fromisoformat(date_sent) if date_sent else None,
            limit=limit,
            page_size=page_size,
        )

        return json.loads(
            json.dumps(
                [msg._properties for msg in messages], indent=2, cls=Message.Encoder
            )
        )

    @action
    def get_message(self, message_sid: str) -> dict:
        """
        Get the details of a stored message.

        :param message_sid: Message SID to be retrieved.
        :return: Message with its properties - see :meth:`.send_message`.
        """
        msg = self.client.messages(message_sid).fetch()
        return msg._properties

    @action
    def update_message(self, message_sid: str, body: str) -> dict:
        """
        Update/redact the body of a message.

        :param message_sid: Message SID to be updated.
        :param body: New message body.
        :return: Updated message with its properties - see :meth:`.send_message`.
        """
        msg = self.client.messages(message_sid).update(body)
        return msg._properties

    @action
    def delete_message(self, message_sid: str):
        """
        Delete a send/received message.

        :param message_sid: Message SID to be deleted.
        """
        self.client.messages(message_sid).delete()

    @action
    def make_call(
        self,
        twiml: str,
        to: str,
        from_: Optional[str] = None,
        method: Optional[str] = None,
        status_callback: Optional[str] = None,
        status_callback_event: Optional[str] = None,
        status_callback_method: Optional[str] = None,
        fallback_url: Optional[str] = None,
        fallback_method: Optional[str] = None,
        send_digits: Optional[str] = None,
        timeout: Optional[int] = 30,
        record: bool = False,
        recording_channels: Optional[int] = None,
        recording_status_callback: Optional[str] = None,
        recording_status_callback_method: Optional[str] = None,
        recording_status_callback_event: Optional[str] = None,
        sip_auth_username: Optional[str] = None,
        sip_auth_password: Optional[str] = None,
        caller_id: Optional[str] = None,
        call_reason: Optional[str] = None,
    ) -> dict:
        """
        Make an automated phone call from a registered Twilio number.

        :param twiml: TwiML containing the logic to be executed in the call
            (see https://www.twilio.com/docs/voice/twiml).
        :param to: Recipient phone number or address book name.
        :param from_: Registered Twilio phone number that will perform the call
            (default: default configured phone number).
        :param method: HTTP method to use to fetch TwiML if it's provided remotely.
        :param status_callback: The URL that should be called to send status
            information to your application.
        :param status_callback_event: The call progress events to be sent to
            the ``status_callback`` URL.
        :param status_callback_method: HTTP Method to use with status_callback.
        :param fallback_url: Fallback URL in case of error.
        :param fallback_method: HTTP Method to use with fallback_url.
        :param send_digits: The digits to dial after connecting to the number.
        :param timeout: Number of seconds to wait for an answer.
        :param record: Whether to record the call.
        :param recording_channels: The number of channels in the final recording.
        :param recording_status_callback: The URL that we call when the
            recording is available to be accessed.
        :param recording_status_callback_method: The HTTP method to use when
            calling the `recording_status_callback` URL.
        :param recording_status_callback_event: The recording status events
            that will trigger calls to the URL specified in
            `recording_status_callback`
        :param sip_auth_username: The username used to authenticate the caller
            making a SIP call.
        :param sip_auth_password: The password required to authenticate the
            user account specified in `sip_auth_username`.
        :param caller_id: The phone number, SIP address, or Client identifier
            that made this call. Phone numbers are in E.164 format (e.g.,
            +16175551212). SIP addresses are formatted as `name@company.com`.
        :param call_reason: Reason for the call (Branded Calls Beta).
        :return: The call properties and details, as a dictionary. Example:

            .. code-block:: json

               {
                  "account_sid": "ACXXX",
                  "annotation": null,
                  "answered_by": null,
                  "api_version": "2010-04-01",
                  "caller_name": null,
                  "date_created": "Tue, 31 Aug 2010 20:36:28 +0000",
                  "date_updated": "Tue, 31 Aug 2010 20:36:44 +0000",
                  "direction": "inbound",
                  "duration": "15",
                  "end_time": "Tue, 31 Aug 2010 20:36:44 +0000",
                  "forwarded_from": "+141586753093",
                  "from": "+15017122661",
                  "from_formatted": "(501) 712-2661",
                  "group_sid": null,
                  "parent_call_sid": null,
                  "phone_number_sid": "PNXXX",
                  "price": "-0.03000",
                  "price_unit": "USD",
                  "sid": "CAXXX",
                  "start_time": "Tue, 31 Aug 2010 20:36:29 +0000",
                  "status": "completed",
                  "subresource_uris": {
                    "notifications": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Notifications.json",
                    "recordings": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Recordings.json",
                    "feedback": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Feedback.json",
                    "feedback_summaries": "/2010-04-01/Accounts/ACXXX/Calls/FeedbackSummary.json",
                    "payments": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Payments.json"
                  },
                  "to": "+14155551212",
                  "to_formatted": "(415) 555-1212",
                  "trunk_sid": null,
                  "uri": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX.json",
                  "queue_time": "1000"
               }

        """
        to = self.address_book.get(to, to)
        call = self.client.calls.create(
            to=to,
            from_=from_ or self.phone_number,
            twiml=twiml,
            method=method,
            status_callback=status_callback,
            status_callback_event=status_callback_event,
            status_callback_method=status_callback_method,
            fallback_url=fallback_url,
            fallback_method=fallback_method,
            send_digits=send_digits,
            timeout=str(timeout) if timeout else None,
            record=record,
            recording_channels=str(recording_channels) if recording_channels else None,
            recording_status_callback=recording_status_callback,
            recording_status_callback_method=recording_status_callback_method,
            recording_status_callback_event=recording_status_callback_event,
            sip_auth_username=sip_auth_username,
            sip_auth_password=sip_auth_password,
            caller_id=caller_id,
            call_reason=call_reason,
        )

        return call._properties

    @action
    def list_calls(
        self,
        to: Optional[str] = None,
        from_: Optional[str] = None,
        parent_call_sid: Optional[str] = None,
        status: Optional[str] = None,
        start_time_before: Optional[str] = None,
        start_time: Optional[str] = None,
        start_time_after: Optional[str] = None,
        end_time_before: Optional[str] = None,
        end_time: Optional[str] = None,
        end_time_after: Optional[str] = None,
        limit: Optional[int] = None,
        page_size: Optional[int] = None,
    ) -> List[dict]:
        """
        List the calls performed by the account, either the full list or those
        that match some filter.

        :param to: Phone number or Client identifier of calls to include
        :param from_: Phone number or Client identifier to filter `from` on
        :param parent_call_sid: Parent call SID to filter on
        :param  status: The status of the resources to read
        :param start_time_before: Only include calls that started on this date
        :param start_time: Only include calls that started on this date
        :param start_time_after: Only include calls that started on this date
        :param end_time_before: Only include calls that ended on this date
        :param end_time: Only include calls that ended on this date
        :param end_time_after: Only include calls that ended on this date
        :param limit: Upper limit for the number of records to return. list() guarantees
            never to return more than limit.  Default is no limit
        :param page_size: Number of records to fetch per request, when not set will use
            the default value of 50 records.  If no page_size is defined
            but a limit is defined, list() will attempt to read the limit
            with the most efficient page size, i.e. min(limit, 1000)
        :return: A list of dictionaries, each representing the information of a
            call. Example:

                .. code-block:: json

                    [
                       {
                          "account_sid": "ACXXX",
                          "annotation": "billingreferencetag1",
                          "answered_by": "machine_start",
                          "api_version": "2010-04-01",
                          "caller_name": "callerid1",
                          "date_created": "Fri, 18 Oct 2019 17:00:00 +0000",
                          "date_updated": "Fri, 18 Oct 2019 17:01:00 +0000",
                          "direction": "outbound-api",
                          "duration": "4",
                          "end_time": "Fri, 18 Oct 2019 17:03:00 +0000",
                          "forwarded_from": "calledvia1",
                          "from": "+13051416799",
                          "from_formatted": "(305) 141-6799",
                          "group_sid": "GPXXX",
                          "parent_call_sid": "CAXXX",
                          "phone_number_sid": "PNXXX",
                          "price": "-0.200",
                          "price_unit": "USD",
                          "sid": "CAXXX",
                          "start_time": "Fri, 18 Oct 2019 17:02:00 +0000",
                          "status": "completed",
                          "subresource_uris": {
                            "feedback": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Feedback.json",
                            "feedback_summaries": "/2010-04-01/Accounts/ACXXX/Calls/FeedbackSummary.json",
                            "notifications": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Notifications.json",
                            "recordings": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Recordings.json",
                            "payments": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX/Payments.json"
                          },
                          "to": "+13051913581",
                          "to_formatted": "(305) 191-3581",
                          "trunk_sid": "TKXXX",
                          "uri": "/2010-04-01/Accounts/ACXXX/Calls/CAXXX.json",
                          "queue_time": "1000"
                      }
                    ]

        """

        call_list = self.client.calls.list(
            to=to,
            from_=from_,
            parent_call_sid=parent_call_sid,
            status=status,
            start_time_before=datetime.datetime.fromisoformat(start_time_before)
            if start_time_before
            else None,
            start_time=datetime.datetime.fromisoformat(start_time)
            if start_time
            else None,
            start_time_after=datetime.datetime.fromisoformat(start_time_after)
            if start_time_after
            else None,
            end_time_before=datetime.datetime.fromisoformat(end_time_before)
            if end_time_before
            else None,
            end_time=datetime.datetime.fromisoformat(end_time) if end_time else None,
            end_time_after=datetime.datetime.fromisoformat(end_time_after)
            if end_time_after
            else None,
            limit=limit,
            page_size=page_size,
        )

        return json.loads(
            json.dumps(
                [call._properties for call in call_list], indent=2, cls=Message.Encoder
            )
        )


# vim:sw=4:ts=4:et:
